# Help strings for fields
wpd = "<a href=""https://en.wikipedia.org/wiki/BibTeX#Field_types"" target=""_blank"">(Wikipedia)</a>"
cf_descriptions = {
    # CodeHierarchies
    ("cat_hierarchies", "source"): "Data producing source defining the hierarchy. For instance “Eurostat” or “FAO”.",
    ("cat_hierarchies", "hierarchy_group"): "Name of a group of hierarchies. Because hierarchies must be members of a HierarchyGroup, when left blank, the Hierarchy name is passed to the HierarchyGroup. At the same time, this means that the Hierarchy is a Code List which is also a base hierarchy, whose codes can be rearranged/reorganized by Hierarchies referring to them.",
    ("cat_hierarchies", "hierarchy_name"): "Name of a hierarchy. The name must be unique inside the group.",
    ("cat_hierarchies", "level"): "Name of a level in a hierarchy. Codes in a Hierarchy can be grouped in levels, this field allows specifying this.",
    ("cat_hierarchies", "referred_hierarchy"): "Name of the base Hierarchy -Code List- where the code is defined originally. Optional when defining a base Hierarchy -Code List-, mandatory when defining Hierarchical Code List.",
    ("cat_hierarchies", "code"): "The formal name of a node in the hierarchy. This name can be used later in different places: the creation of hierarchical code lists, values in attributes, expressions... .",
    ("cat_hierarchies", "parent_code"): "If the code being defined does not have parent the value is empty, if not, specify the parent Code.",
    ("cat_hierarchies", "label"): "A label that will be used when elaborating datasets or visualizations.",
    ("cat_hierarchies", "description"): "A free description optionally used for readability.",
    ("cat_hierarchies", "expression"): "A simple arithmetic expression referring to codes in the same hierarchy. No circular references allowed.",
    ("cat_hierarchies", "geolocation_ref"): "The ID of a 'geographic dataset' reference.",
    ("cat_hierarchies", "geolocation_code"): "When “Code” does not identify a feature in the referred dataset, this field can be used. ",
    ("cat_hierarchies", "attributes"): "A list of freely definable attributes attached to the code. Columns with a header starting in “@” are considered attribute columns that are appended to the list of attributes.",
    # CodeHierarchiesMapping
    ("cat_hier_mapping", "source_dataset"): "Dataset containing the definition of the hierarchy (for hierarchies defined through importing datasets).",
    ("cat_hier_mapping", "source_hierarchy"): "Name of the origin Hierarchy.",
    ("cat_hier_mapping", "source_code"): "Name of a code in the origin Hierarchy.",
    ("cat_hier_mapping", "destination_hierarchy"): "Name of the destination Hierarchy.",
    ("cat_hier_mapping", "destination_code"): "Name of a code in the destination Hierarchy.",
    ("cat_hier_mapping", "weight"): "Weight permits splitting an origin code in more than one destination codes, or simply change the scale of an origin code into the destination code. If a code is split the sum should not be more than one. Expressions using parameters are permitted. In case “weight” is left empty, it is equivalent to specifying “1”.",
    # DatasetDef
    ("datasetdef", "dataset_name"): "Name of the custom dataset.",
    ("datasetdef", "dataset_data_location"): "Location of data of this dataset. If data is embedded in the same spreadsheet (DatasetData command), “data://” must be used. If not, the location of a publicly available CSV or XLSX file must be provided, using a URL. If a XLSX file is used, the worksheet name to be used is specified appending to the URL the character “#” and the name of the worksheet. If the worksheet is local, the syntax “data://#” followed by the “DatasetData” worksheet name can be used.",
    ("datasetdef", "concept_type"): "Empty for the Dataset header row, One of the allowed values to define a concept.",
    ("datasetdef", "concept_name"): "Name of the Concept being defined.",
    ("datasetdef", "concept_data_type"): "Data-type of the Concept.",
    ("datasetdef", "concept_domain"): "Allows constraining the values of a concept. Dimensions typically are “Code”, so a Hierarchy name must be specified. Measures are normally “Numbers”, here an interval could be stated. Mandatory if “ConceptDataType” is “Code”.",
    ("datasetdef", "concept_description"): "Description of the dataset (first appearance of the dataset) or Concept (following rows).",
    ("datasetdef", "attributes"): "A list of freely definable attributes attached to the dataset or concept.",
    # Parameters
    ("parameters", "name"): "Name of the parameter. Must be unique across the case study.",
    ("parameters", "type"): "Type of the parameter.",
    ("parameters", "domain"): "Set of potential values for the parameter. Mandatory for “Code”, an optional range for “Number”.",
    ("parameters", "value"): "Either a literal value or a conditional list of expressions which can have other parameters (declared previously). An empty value is permitted because “Solver” command can define -or redefine- parameters.",
    ("parameters", "group"): "Name of the group of parameters to which the current parameter is a member of, just for readability.",
    ("parameters", "description"): "A description of the meaning and intent of the parameter.",
    ("parameters", "attributes"): "A list of freely definable attributes attached to the parameter. Columns with a header starting in “@” are considered attribute columns that are appended to the list of attributes.",
    # InterfaceTypes
    ("interface_types", "interface_type_hierarchy"): "Name of the hierarchy to which the InterfaceType is a member.",
    ("interface_types", "interface_type"): "Name of the InterfaceType. Unique among all Interface Types.",
    ("interface_types", "sphere"): "The default value of sphere, either “Technosphere” or “Biosphere”, for Interfaces of this type (can be overridden by each Interface when it is declared). Indicates the type of process originating the input flow. Or the destination for output flows.",
    ("interface_types", "roegen_type"): "The default Either “Flow” or “Fund”, indicates the type of factor for the interface in accordance with Roegen theory.",
    ("interface_types", "parent_interface_type"): "Name of the parent InterfaceType, if any.",
    ("interface_types", "level"): "Semantic scale level at which the interface type is categorized.",
    ("interface_types", "formula"): "A formula combining arithmetically InterfaceTypes in different branches of the same hierarchy.",
    ("interface_types", "description"): "Description of the InterfaceType.",
    ("interface_types", "unit"): "The default unit for quantities attached to the InterfaceType. A very large list of units is recognized.",
    ("interface_types", "opposite_processor_type"): "To which of the case study contexts, “Local”, “External”, “Environment” or “ExternalEnvironment”, does the Processor opposite to where Interfaces of the InterfaceType pertain to.",
    ("interface_types", "attributes"): "A list of freely definable attributes attached to an InterfaceType. Columns with a header starting in “@” are considered attribute columns that are appended to the list of attributes.",
    # Processors
    ("processors", "processor_group"): "A group to which the Processor will be a member of.",
    ("processors", "processor"): "Name of the Processor.",
    ("processors", "parent_processor"): "Name of the parent Processor regarding “part-of” relation. It is a quick way of specifying a hierarchy of functional composition. Another way is through the command “Relationships”.",
    ("processors", "parent_processor_weight"): "Contribution of the processor to its parent when aggregating. 1 by default, it may be useful for multifunctional processors.",
    ("processors", "behave_as_processor"): "Make current processor to seem as the named processor when aggregating 'part-of' to the current ParentProcessor. A different processor could be used for another parent.",
    ("processors", "subsystem_type"): "The main context of the Processor. Default value 'Local'.",
    ("processors", "processor_system"): "The system to which the processor is member of. It is allowed to connect processors from multiple systems (the initial intent is to label a system as a country, like “ES”, “NL”, “IT”). Default value: 'default'.",
    ("processors", "functional_or_structural"): "One of two options must be specified: Functional or Structural. “Functional” means the Processor is a function, which is the goal in the abstract, opposed to “Structure” which is a Processor representing a materialized goal. Structural Processors should be part-of functional Processors (but structural Processors can have other structural Processors inside, and functional Processors can have functional Processors inside).",
    ("processors", "instance_or_archetype"): "Indicates if the processor is accounted (Yes or Instance) or (No or Archetype) if it is being defined to be used as prototype or template (for instance intensive -or unit- processors that will be copied -cloned- and scaled up), and must stay outside of the accounting process.",
    ("processors", "level"): "Semantic scale level the processor is located into.",
    ("processors", "stock"): "Inform that the Processor is a stock (“Yes”) or a producer/consumer (“No”).",
    ("processors", "description"): "Description of the Processor.",
    ("processors", "geolocation_ref"): "The ID of a “geographic dataset” reference.",
    ("processors", "geolocation_code"): "Either a Code in a Hierarchy having a geographic object attached to it, or the code of an object inside the spatial dataset specified in “GeolocationRef”.",
    ("processors", "geolocation_latlong"): "The Latitude and Longitude of a representative point of the Processor (using WGS84 projection).",
    ("processors", "attributes"): "A list of freely definable attributes attached to a Processor. Columns with a header starting in “@” are considered attribute columns that are appended to the list of attributes.",
    # Interfaces
    ("interfaces_and_qq", "processor"): "Processor to which the Interface is added. Or the quantity, if the Interface has been already added.",
    ("interfaces_and_qq", "interface_type"): "Name of the Interface Type to be used.",
    ("interfaces_and_qq", "interface"): "Name of the Interface in case it is needed because by default the InterfaceType name is used for the Interface (if two Interfaces with the same InterfaceType are needed in the same Processor, they must have different names).",
    ("interfaces_and_qq", "sphere"): "The sphere, either “Technosphere” or “Biosphere”, for the Interface being defined, potentially overriding the equivalent property of InterfaceType. Indicates the type of process (artificial or natural) originating the input flow. Or the destination for output flows.",
    ("interfaces_and_qq", "roegen_type"): "Either “Flow” or “Fund”, indicates the type of factor for the Interface in accordance with Roegen’s theory. Overrides the RoegenType of the InterfaceType if it was specified.",
    ("interfaces_and_qq", "orientation"): "Orientation, “Input” or “Output”, of the Interface. Overrides the Orientation of the InterfaceType if it was specified.",
    ("interfaces_and_qq", "opposite_processor_type"): "Used to point which of the case study compartments do Processors opposite to the Interface pertain to.",
    ("interfaces_and_qq", "geolocation_ref"): "Reference to information on where the Interface is located.",
    ("interfaces_and_qq", "geolocation_code"): "Either a Code in a Hierarchy having a geographic object attached to it, or the code of an object inside the spatial dataset specified in “GeolocationRef”.",
    ("interfaces_and_qq", "range"): "Numeric interval where the value for the interface is acceptable. Useful for constraints in automated solution finding (optimization and other kinds).",
    ("interfaces_and_qq", "range_unit"): "Unit for the numbers in the 'Range' field",
    ("interfaces_and_qq", "interface_attributes"): "A list of freely definable attributes attached to an Interface. Columns with a header starting in “I@” are considered attribute columns that are appended to the list of attributes of the Interface. Quantities have a separate list of attributes, see QuantityAttributes.",
    ("interfaces_and_qq", "value"): "An arithmetic-boolean expression with parameters.",
    ("interfaces_and_qq", "unit"): "Unit for Value. A very large list of units is recognized. Overrides the Unit of the InterfaceType, if it was specified.",
    ("interfaces_and_qq", "relative_to"): "If used, it would be the name of an Interface defining the unit to scale the currently stated Value. For instance, if the Unit is “kg” and “RelativeTo” is “LU” in hectares, once “LU” is quantified, lets assume in “10 ha”, the resulting amount would be multiplied by 10, and the unit would remain to be “kg”.",
    ("interfaces_and_qq", "uncertainty"): "Statistical characterization of the uncertainty of Value. This field should be elaborated together with solvers capable of exploiting the uncertainty information: sensitivity analysis, observability analysis (recognize the inherent strength-weakness of assessment including this measure).",
    ("interfaces_and_qq", "assessment"): "Assessment according to NUSAP. This should be elaborated together with solvers capable of symbolically analyzing categorizations of assessment.",
    ("interfaces_and_qq", "pedigree_matrix"): "Reference to a previously declared PedigreeMatrix (“NUSAP.PM” command).",
    ("interfaces_and_qq", "pedigree"): "Codes in the PedigreeMatrix encoding qualities of the diagnostic quantity.",
    ("interfaces_and_qq", "time"): "To which time period the quantity corresponds to.",
    ("interfaces_and_qq", "qq_source"): "Reference to a source defining how (provenance) or from (bibliography) the quantity was obtained.",
    ("interfaces_and_qq", "number_attributes"): "A list of freely definable attributes attached to a Quantity. Columns with a header starting in “N@” are considered attribute columns that are appended to the list of attributes of the Quantity.",
    ("interfaces_and_qq", "comments"): "Free comments attached to the quantity.",
    # Relationships
    ("relationships", "source_processor"): "An expression resolving to a set of processors that will be in the origin part of the relations (one or more) to be specified. If “Origin” field is not specified, this field is mandatory.",
    ("relationships", "source_interface"): "The name of an Interface present in all the Processors in the set of origin Processors. When connecting Processors it can be empty. Also, when connecting Interfaces, if the destination Interface is specified, an equally named Interface is assumed in the origin Processor.",
    ("relationships", "target_processor"): "An expression resolving to a set of processors that will be in the destination part of the relations (one or more) to be specified. If “Destination” field is not specified, this field is mandatory.",
    ("relationships", "target_interface"): "The name of an Interface present in all the Processors in the set of destination Processors. When connecting Processors it can be empty. Also, when connecting Interfaces, if the origin Interface is specified, an equally named Interface is assumed in the destination Processor.",
    ("relationships", "back_interface"): "The name of an Interface in the origin processor, used if a change of scale is requested, allowing to account back to the origin processor the amount. It is intended for expressing the accounting of money obtained in exchange for some good.",
    ("relationships", "relation_type"): ["Please, refer to MAGIC - ""<a href=""https://docs.google.com/document/d/1kGJI7m9fL9eRHEG4CICzOCQM5qOBdFZ3eDR43d1dME8/edit#heading=h.vqpx27xubbp8"" target=""_blank"">DMP Annex 3</a>"" (text too long to show here).",
                                         "One of:<br><ul>"
                                         "<li>PartOf or “|”. Origin Processor is “part-of” (child) Destination Processor (parent). No assumptions on flows between child and parent. It is worth stressing it is “functionally part-of” more than “materially part-of”, which is not so interesting for MuSIASEM analysis.</li>"
                                         "<li>Flow or “>” (“<” for reverse flow). A flow from an Interface in the Origin Processor to an Interface in the Destination Processor. Both Interfaces must be of the same InterfaceType. If the flow is between two Interfaces from Ancestor-Descendant Processors, the Orientation of both must be the same and always from Output to Output. If it is between siblings, Orientations must be the opposite, and always from Output to Input.</li>"
                                         "<li>Scale. An information flow from an Interface into another, meaning that the Destination Interface is equal to the Origin Interface times the Weight.</li>"
                                         "<li>IsA. State that origin “is-a” type of destination (equivalent function).</li>"
                                         "<li>AsA. To declare that view (mask) origin can be viewed as destination (also equivalent function; “is-a” is declared, “as-a” is dynamic).</li>"
                                         "<li>Associate. State that Origin Processor is associated with Destination Processor. The association has to be completed with Origin and Destination Cardinality fields. In comments it is recommended to state the roles of both Origin and Destination Processor. This relation can also be used to avoid flows between Processors not “part-of” Processors involved in “associate” relations.</li>"
                                         "</ul>"
                                         ],
    ("relationships", "flow_weight"): "Proportion of the magnitude of the Origin Interface that goes into the Destination Interface.",
    ("relationships", "change_type_scale"): "In case two Interfaces of different InterfaceTypes are connected by a flow (exchange), create an ad-hoc line of a ScaleChangeMap.",
    ("relationships", "source_cardinality"): "Only for “Associate” relations.",
    ("relationships", "target_cardinality"): "Only for “Associate” relations.",
    ("relationships", "attributes"): "A list of freely definable attributes attached to a Relationship. Columns with a header starting in “@” are considered attribute columns that are appended to the list of attributes.",
    # ProcessorScalings
    ("processor_scalings", "invoking_processor"): "Parent processor for hierarchical scales, origin sibling processor for sequential scales.",
    ("processor_scalings", "requested_processor"): "Child processor for hierarchical scales, destination processor for sequential scales. Depending on the ScalingType, this Processor will exist previously or it will be created from a template processor (unit -or intensive- processor).",
    ("processor_scalings", "scaling_type"): ["Please, refer to MAGIC - ""<a href=""https://docs.google.com/document/d/1kGJI7m9fL9eRHEG4CICzOCQM5qOBdFZ3eDR43d1dME8/edit#heading=h.l6gqwwmmahkc"" target=""_blank"">DMP Annex 3</a>"" (text too long to show here).",
                                             "Defines the operations requested for the process of scaling:<br><ul>"
                                             "<li>“CloneAndScale”. Clones “RequestedProcessor” as a child of “InvokingProcessor”; second, constraints the value of “RequestedInterface” to the value of “InvokingInterface”, scaled by “Scale”. “RequestedProcessor” must be an archetype (opposed to instance) “unit processor”.</li>"
                                             "<li>“Scale”. Constraints the value of “RequestedInterface” to the value of “InvokingInterface”, scaled by “Scale”. “RequestedProcessor” must be an instance (opposed to archetype) “unit processor”.</li>"
                                             "<li>“Clone”. Clones “RequestedProcessor” as a child of “InvokingProcessor”. It does not apply a “Scale” relationship, which should be applied later, either using ProcessorScalings command and ScalingType column equal to “Scale” or Relationships command with RelationType command equal to “Scale”.</li>"
                                             "<li>“CloneScaled”. Clones “RequestedProcessor” as a child of “InvokingProcessor”; second, scales the new processor using “Scale” as the value of “RequestedInterface”. “RequestedProcessor” must be an archetype (opposed to instance) “unit processor”.</li>"
                                             "</ul>"
                                             ],
    ("processor_scalings", "invoking_interface"): "Interface to be used for the invoking processor.",
    ("processor_scalings", "requested_interface"): "Interface to be used for the requested processor.",
    ("processor_scalings", "scale"): "Factor by which invoking interface is multiplied to define the requested interface.",
    ("processor_scalings", "new_processor_name"): "If a new processor is created (“CloneAndScale” is used) the name of the processor is in this field. If not, the name of the requested processor is used.",
    ("processor_scalings", "processor_group"): "If a new processor is created ('CloneAndScale' is used) the group to which the processor is added.",
    ("processor_scalings", "parent_processor"): "If a new processor is created ('CloneAndScale' is used) the parent of the new processor.",
    ("processor_scalings", "subsystem_type"): "If a new processor is created ('CloneAndScale' is used) the context of the new Processor.",
    ("processor_scalings", "processor_system"): "If a new processor is created ('CloneAndScale' is used), the system to which the processor is member of.",
    ("processor_scalings", "level"): "If a new processor is created ('CloneAndScale' is used), semantic scale level the processor is located into.",
    ("processor_scalings", "description"): "If a new processor is created ('CloneAndScale' is used), the description attached to the new processor.",
    ("processor_scalings", "geolocation_ref"): "If a new processor is created ('CloneAndScale' is used), ID of a geolocation reference for this new processor.",
    ("processor_scalings", "geolocation_code"): "If a new processor is created ('CloneAndScale' is used), Code in a hierarchy or geographic reference.",
    ("processor_scalings", "attributes"): "If a new processor is created ('CloneAndScale' is used), attributes for the new processor.",
    # ScaleChangeMap
    ("scale_conversion_v2", "source_hierarchy"): "Name of the origin InterfaceTypes Hierarchy. It is optional if the origin InterfaceType name is unique among InterfaceTypes. Otherwise, it must be specified.",
    ("scale_conversion_v2", "source_interface_type"): "Name of the InterfaceType in the origin InterfaceTypes Hierarchy.",
    ("scale_conversion_v2", "target_hierarchy"): "Name of the destination InterfaceTypes Hierarchy. It is optional if the destination InterfaceType name is unique among InterfaceTypes. Otherwise, it must be specified.",
    ("scale_conversion_v2", "target_interface_type"): "Name of a code in the destination InterfaceTypes Hierarchy.",
    ("scale_conversion_v2", "source_context"): "Specification of which Processors are to be used to match the Origin part of the context.",
    ("scale_conversion_v2", "target_context"): "Specification of which Processors are to be used to match the Destination part of the context.",
    ("scale_conversion_v2", "scale"): "Linear scale transform from origin to destination. Both origin and destination can appear multiple times, so the multiplicity potentially can be many to many.",
    ("scale_conversion_v2", "source_unit"): "Specification of unit for which the scale factor is valid. If quantities for Interfaces are expressed in other physically equivalent unit, automatic unit conversion will be performed before the scaling.",
    ("scale_conversion_v2", "target_unit"): "Specification of unit (destination) for which the scale factor is valid. If quantities for Interfaces are expressed in other physically equivalent unit, automatic unit conversion will be performed before the scaling.",
    # ScalarBenchmarks
    ("scalar_indicator_benchmarks", "benchmark_group"):
    "To which of the set of predefined groups this benchmark is member of.",
    ("scalar_indicator_benchmarks", "stakeholders"):
    "A list of names that can be used to frame which stakeholders consider/use this benchmark.",
    ("scalar_indicator_benchmarks", "benchmark"): "Name of the benchmark.",
    ("scalar_indicator_benchmarks", "range"): "Numeric interval capturing one of the categories of the benchmark.",
    ("scalar_indicator_benchmarks", "unit"): "The unit under which the range is expressed.",
    ("scalar_indicator_benchmarks", "category"):
    "A relatively formal name that can be used later to elaborate visualizations (color, icon, ...)",
    ("scalar_indicator_benchmarks", "label"): "A short string sketching the level or category inside the benchmark.",
    ("scalar_indicator_benchmarks", "description"): "Description of the category or the benchmark.",
    # ScalarIndicators
    ("scalar_indicators", "indicator_name"): "Name for the indicator, will appear in result matrices",
    ("scalar_indicators", "local"):
    "'Yes' if the formula for the indicator uses values attached to a single processor. 'No' if it is a formula "
    "referring to values in the whole model. Only 'Yes' allowed currently.",
    ("scalar_indicators", "formula"):
    "Arithmetic expression to calculate the indicator. If it is local, it can refer to interface names and also to "
    "scenario parameters.",
    ("scalar_indicators", "benchmarks"): "Name of benchmarks under which the indicator is framed.",
    ("scalar_indicators", "description"): "Meaning of the indicator.",
    # MatrixIndicators
    ("matrix_indicators", "indicator_name"): "Name of the matrix indicator.",
    ("matrix_indicators", "scope"): "Used to enable quantifying openness.",
    ("matrix_indicators", "processors_selector"): "Used to obtain a set of processors for the rows of the matrix.",
    ("matrix_indicators", "interfaces_selector"): "Interfaces to place in the columns of the matrix.",
    ("matrix_indicators", "indicators_selector"): "Local scalar indicators to place in the columns of the matrix.",
    ("matrix_indicators", "attributes_selector"): "Attributes (from processor) to place in the columns of the matrix.",
    ("matrix_indicators", "description"): "Intent of the indicator.",
    # ProblemStatement
    ("problem_statement", "scenario_name"): "Name of the scenario.",
    ("problem_statement", "parameter"): "Name of the parameter that is being modified.",
    ("problem_statement", "parameter_value"): "",
    ("problem_statement", "description"): "Explanation of the value or the intent of the scenario.",
    # ScaleChangeMap
    ("scale_conversion_v2", "source_hierarchy"): "Name of the origin InterfaceTypes Hierarchy. It is optional if the origin InterfaceType name is unique among InterfaceTypes. Otherwise, it must be specified.",
    ("scale_conversion_v2", "source_interface_type"): "Name of the InterfaceType in the origin InterfaceTypes Hierarchy.",
    ("scale_conversion_v2", "target_hierarchy"): "Name of the destination InterfaceTypes Hierarchy. It is optional if the destination InterfaceType name is unique among InterfaceTypes. Otherwise, it must be specified.",
    ("scale_conversion_v2", "target_interface_type"): "Name of a code in the destination InterfaceTypes Hierarchy.",
    ("scale_conversion_v2", "source_context"): "Specification of which Processors are to be used to match the Origin part of the context.",
    ("scale_conversion_v2", "target_context"): "Specification of which Processors are to be used to match the Destination part of the context.",
    ("scale_conversion_v2", "scale"): "Linear scale transform from origin to destination. Both origin and destination can appear multiple times, so the multiplicity potentially can be many to many.",
    ("scale_conversion_v2", "source_unit"): "Specification of unit for which the scale factor is valid. If quantities for Interfaces are expressed in other physically equivalent unit, automatic unit conversion will be performed before the scaling.",
    ("scale_conversion_v2", "target_unit"): "Specification of unit (destination) for which the scale factor is valid. If quantities for Interfaces are expressed in other physically equivalent unit, automatic unit conversion will be performed before the scaling.",
    # ImportCommands
    ("import_commands", "workbook_name"): "URL of the workbook containing the commands to be reused.",
    ("import_commands", "worksheets"): "Comma separated list of worksheets in the workbook to import. If not specified, all commands from the specified Workbook will be imported.",
    # ListOfCommands
    ("list_of_commands", "worksheet"): "Name of the worksheet.",
    ("list_of_commands", "command"): "Name of one of the supported commands.",
    ("list_of_commands", "comment"): "Description of the content of the worksheet.",
    # RefGeographic
    ("ref_geographical", "ref_id"): "A unique reference name.",
    ("ref_geographical", "data_location"): "Location of a vector layer in a supported format (GeoJSON currently).",
    ("ref_geographical", "title"): "Title.",
    ("ref_geographical", "description"): "Description.",
    ("ref_geographical", "bounding_box"): "Extents of the dataset expressed in: west and east longitudes, south and north latitudes (WGS84), using decimal degrees with at least two decimals.",
    ("ref_geographical", "topic_category"): 'It is a free string, although one or more of the following should be used: "Farming", "Biota", "Boundaries", "Climatology", "Meteorology", "Atmosphere", "Economy", "Elevation", "Environment", "GeoscientificInformation", "Health", "Imagery", "BaseMaps", "EarthCover", "Intelligence", "Military", "InlandWaters", "Location", "Oceans", "Planning", "Cadastre", "Society", "Structure", "Transportation", "Utilities", "Communication".',
    ("ref_geographical", "temporal_extent"): "Temporal period covered by the layer.",
    ("ref_geographical", "point_of_contact"): "Address of responsible organization/person.",
    ("ref_geographical", "type"): "Just 'dataset' allowed.",
    # RefBibliographic
    ("ref_bibliographic", "ref_id"): "A unique name for the reference.",
    ("ref_bibliographic", "bib_file_url"): "URL of a BibTeX file containing one or more bibliographic references. The RefID must match one of the references in this file.",
    ("ref_bibliographic", "entry_type"): f"The type of bibliographic reference {wpd}.",
    ("ref_bibliographic", "address"): f"Publisher's address (usually just the city, but can be the full address for lesser-known publishers) {wpd}.",
    ("ref_bibliographic", "annote"): f"An annotation for annotated bibliography styles (not typical) {wpd}.",
    ("ref_bibliographic", "author"): f"The name(s) of the author(s) (in the case of more than one author, separated by and) {wpd}.",
    ("ref_bibliographic", "booktitle"): f"The title of the book, if only part of it is being cited {wpd}.",
    ("ref_bibliographic", "chapter"): f"The chapter number {wpd}.",
    ("ref_bibliographic", "crossref"): f"The key of the cross-referenced entry {wpd}.",
    ("ref_bibliographic", "edition"): f'The edition of a book, long form (such as "First" or "Second") {wpd}.',
    ("ref_bibliographic", "editor"): f"The name(s) of the editor(s) {wpd}.",
    ("ref_bibliographic", "how_published"): f"How it was published, if the publishing method is nonstandard {wpd}.",
    ("ref_bibliographic", "institution"): f"The institution that was involved in the publishing, but not necessarily the publisher {wpd}.",
    ("ref_bibliographic", "journal"): f"The journal or magazine the work was published in {wpd}.",
    ("ref_bibliographic", "key"): f'A hidden field used for specifying or overriding the alphabetical order of entries (when the "author" and "editor" fields are missing). Note that this is very different from the key (mentioned just after this list) that is used to cite or cross-reference the entry. {wpd}.',
    ("ref_bibliographic", "month"): f"The month of publication (or, if unpublished, the month of creation) {wpd}.",
    ("ref_bibliographic", "note"): f"Miscellaneous extra information {wpd}.",
    ("ref_bibliographic", "number"): f'The "(issue) number" of a journal, magazine, or tech-report, if applicable. Note that this is not the "article number" assigned by some journals. {wpd}.',
    ("ref_bibliographic", "organization"): f"The conference sponsor {wpd}.",
    ("ref_bibliographic", "pages"): f"Page numbers, separated either by commas or double-hyphens. {wpd}.",
    ("ref_bibliographic", "publisher"): f"The publisher's name {wpd}.",
    ("ref_bibliographic", "school"): f"The school where the thesis was written {wpd}.",
    ("ref_bibliographic", "series"): f"The series of books the book was published in {wpd}.",
    ("ref_bibliographic", "title"): f"The title of the work {wpd}.",
    ("ref_bibliographic", "type"): f'The field overriding the default type of publication (e.g. "Research Note" for techreport, "{{PhD}} dissertation" for phdthesis, "Section" for inbook/incollection) {wpd}.',
    # ("ref_bibliographic", "url"): f"",
    ("ref_bibliographic", "volume"): f"The volume of a journal or multi-volume book {wpd}.",
    ("ref_bibliographic", "year"): f"The year of publication (or, if unpublished, the year of creation) {wpd}.",
    # RefProvenance
    ("ref_provenance", "ref_id"): "A unique name for the reference.",
    ("ref_provenance", "provenance_file_url"): "URL of a file containing a PROV-XML full description of the provenance.",
    # ("ref_provenance", "agent_type"): "",
    ("ref_provenance", "agent"): "A description of which is the main agent (something able to perform activities).",
    ("ref_provenance", "activities"): "A description of the activities (processes) carried out.",
    ("ref_provenance", "entities"): "A description of the entities (information objects) used to elaborate the information.",
    # DatasetQry
    ("datasetqry", "inputdataset"): "Name of the dataset used as input (only one dataset can be specified, the first in the column, others are ignored).",
    ("datasetqry", "availableatdatetime"): "Date and time of the version preceding the specified date. (Currently not considered).",
    ("datasetqry", "starttime"): "Starting year.",
    ("datasetqry", "endtime"): "Ending year.",
    ("datasetqry", "resultdimensions"): "Below this column header, a series of dimension names can be enumerated, which will not appear in the resulting data cube grouping values using the aggregation function.",
    ("datasetqry", "resultmeasures"): "A list of measures to include in the resulting data cube. Measure names can appear multiple times. If empty, ALL measures are included.",
    ("datasetqry", "resultmeasuresaggregation"): 'The functions recognized are: "sum", "avg", "count", "sumna" (sum ignoring not available), "countav" (count only defined elements), "avgna", "pctna" (percentage of not available with regard to the total number of elements -available or not-). Measures and aggregation functions form pairs (measure- aggregation function).',
    ("datasetqry", "resultmeasurename"): "For each measure, an alias with the name of the output column can be specified. If not, the output column is a composition of the measure name and the aggregation function.",
    ("datasetqry", "outputdataset"): "Name of the dataset variable that will contain the result. This dataset will be available internally (<a href=""https://docs.google.com/document/d/1kGJI7m9fL9eRHEG4CICzOCQM5qOBdFZ3eDR43d1dME8/edit#heading=h.z8oasdu66rh"" target=""_blank"">dataset expansion</a>) and as output dataset (to download it).",
}



# -*- coding: utf-8 -*-
"""
* Registry of objects. Add, remove, search
* Support for high level operations: directly create and/or modify objects, calling the specification API. Create connections
"""
import copy
import datetime
import json
import logging
import uuid
from enum import Enum
from typing import List, Union, Dict, NoReturn

import pandas as pd

import nexinfosys
from nexinfosys.command_generators import Issue, IssueLocation, IType
from nexinfosys.command_generators.parsers_factory import commands_container_parser_factory
from nexinfosys.common.helper import create_dictionary, strcmp
from nexinfosys.initialization import prepare_and_reset_database_for_tests
from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects
from nexinfosys.model_services import State
from nexinfosys.models.musiasem_concepts import ProblemStatement, FactorsRelationDirectedFlowObservation, Processor, \
    Factor, Parameter, FactorInProcessorType
from nexinfosys.models.musiasem_methodology_support import (User,
                                                            CaseStudy,
                                                            CaseStudyVersion,
                                                            CaseStudyVersionSession,
                                                            CommandsContainer,
                                                            force_load,
                                                            DBSession)
from nexinfosys.serialization import serialize_state, deserialize_state
from nexinfosys.solving import BasicQuery
from nexinfosys.solving.flow_graph_solver import flow_graph_solver, evaluate_parameters_for_scenario, get_dataset

logger = logging.getLogger(__name__)


class Identity:
    pass

# class IdentityCredentials:
#     pass
#
# class NISystem:
#     def __init__(self):
#         self._authentication_service = None
#         self._authorization_service = None
#         self._base_url = None
#         self._configuration_manager = None
#         self._plugin_manager = None
#
#
#     def gather_credentials(self) -> IdentityCredentials:
#         return IdentityCredentials()
#
#     def login(self, ic: IdentityCredentials):
#         # Check credentials using authentication service
#         return Identity()
#
#     def logout(self, id: Identity):
#         pass


class CommandResult:
    pass

# class SessionCreationAction(Enum):  # Used in FlowFund
#     """
#
#         +--------+--------+---------+---------------------------------+
#         | branch | clone  | restart |        Behavior                 |
#         +--------+--------+---------+---------------------------------+
#         | True   | True   | True    | Branch & New ReproducibleSession|
#         | True   | True   | False   | Branch & Clone                  |
#         | True   | False  | True    | Branch & New ReproducibleSession|
#         | True   | False  | False   | Branch & New ReproducibleSession|
#         | False  | True   | True    | New CS (not CS version)         |
#         | False  | True   | False   | New CS & Clone                  |
#         | False  | False  | True    | Restart                         |
#         | False  | False  | False   | Continue                        |
#         +--------+--------+---------+---------------------------------+
#     """
#     BranchAndNewWS = 7
#     BranchAndCloneWS = 6
#     NewCaseStudy = 3
#     NewCaseStudyCopyFrom = 2
#     Restart = 1
#     Continue = 0
# ---------------------------------------------------------------------------------------------------------------------

# #####################################################################################################################
# >>>> PRIMITIVE COMMAND & COMMAND GENERATOR PROCESSING FUNCTIONS <<<<
# #####################################################################################################################


def executable_command_to_commands_container(e_cmd: IExecutableCommand):
    """
    IExecutableCommand -> CommandsContainer

    The resulting CommandsContainer will be always in native (JSON) format, because the specification
    to construct an IExecutableCommand has been translated to this native format.

    :param command:
    :return:
    """
    d = {"command": e_cmd._serialization_type,
         "label": e_cmd._serialization_label,
         "content": e_cmd.json_serialize()}

    return CommandsContainer.create("native", "application/json", json.dumps(d).encode("utf-8"))


def persistable_to_executable_command(p_cmd: CommandsContainer, limit=1000):
    """
    A persistable command can be either a single command or a sequence of commands (like a spreadsheet). In the future
    it could even be a full script.

    Because the executable command is DIRECTLY executable, it is not possible to convert from persistable to a single
    executable command. But it is possible to obtain a list of executable commands, and this is the aim of this function

    The size of the list can be limited by the parameter "limit". "0" is for unlimited

    :param p_cmd:
    :return: A list of IExecutableCommand
    """
    # Create commands generator from factory (from generator_type and file_type)
    state = State()
    commands_generator = commands_container_parser_factory(p_cmd.generator_type, p_cmd.file_type, p_cmd.file, state)

    # Loop over the IExecutableCommand instances
    issues_aggreg = []
    outputs = []
    count = 0
    for cmd, issues in commands_generator:
        # If there are syntax ERRORS, STOP!!!
        stop = False
        if issues and len(issues) > 0:
            for t in issues:
                if t[0] == 3:  # Error
                    stop = True
        if stop:
            break

        issues_aggreg.extend(issues)

        count += 1
        if count >= limit:
            break


def execute_command(state, e_cmd: "IExecutableCommand") -> nexinfosys.IssuesOutputPairType:
    if e_cmd:
        return e_cmd.execute(state)
    else:
        return [], None  # issues, output


def execute_command_container(state, p_cmd: CommandsContainer, ignore_imports=False):
    return execute_command_container_file(state, p_cmd.generator_type, p_cmd.content_type, p_cmd.content, ignore_imports)


def execute_command_container_file(state, generator_type, file_type: str, file, ignore_imports):
    """
    This could be considered the MAIN method of the processing.
    1) Assuming an initial "state" (that can be clean or not),
    2) loops over the commands represented in a "file" of some of the supported types (JSON, Spreadsheet)
       2.1) Parses each command, returning an IExecutableCommand instance (containing the JSON definition inside)
       2.2) Executes each command, if there are no error issues. Command execution can reads and modifies "state"

    :param generator_type: Which commands generator (parser + constructor of IExecutableCommand instances) is used
    :param file_type: The file format
    :param file: The file contents
    :return: Issues and outputs (no outputs still required, probably won't be needed)
    """
    # Create commands generator from factory (from generator_type and file_type)
    commands_generator = commands_container_parser_factory(generator_type, file_type, file, state, None, None, ignore_imports)

    # Loop over the IExecutableCommand instances
    issues_aggreg = []
    outputs = []
    cmd_number = 0
    for cmd, issues in commands_generator:
        if len(issues) > 0:
            new_issues, errors_exist = transform_issues(issues, cmd, cmd_number)
            c = "\n"
            logging.debug(f"Issues:\n{c.join([i.description for i in new_issues])}")
        else:
            logging.debug(f"{type(cmd)} {cmd._source_block_name if hasattr(cmd, '_source_block_name') else ''}; # syntax issues: {len(issues)}")
        cmd_number += 1  # Command counter

        errors_exist = False

        if issues and len(issues) > 0:
            new_issues, errors_exist = transform_issues(issues, cmd, cmd_number)
            issues_aggreg.extend(new_issues)

        if errors_exist:
            break

        # ## COMMAND EXECUTION ## #
        issues, output = execute_command(state, cmd)

        if issues and len(issues) > 0:
            new_issues, errors_exist = transform_issues(issues, cmd, cmd_number)
            issues_aggreg.extend(new_issues)

        if output:
            outputs.append(output)

        if errors_exist:
            break

    return issues_aggreg, outputs


def transform_issues(issues: List[Union[dict, nexinfosys.Issue, tuple, Issue]], cmd, sheet_number: int) -> (List[Issue], bool):

    errors_exist = False
    new_issues: List[Issue] = []

    for i in issues:
        if isinstance(i, dict):
            issue = Issue(itype=IType(i["type"]), description=i["message"], ctype=i["c_type"],
                          location=IssueLocation(sheet_name=i["sheet_name"], sheet_number=i["sheet_number"]))
        elif isinstance(i, nexinfosys.Issue):  # namedtuple
            issue = Issue(itype=i.type, description=i.message, ctype=i.c_type,
                          location=IssueLocation(sheet_name=i.sheet_name, sheet_number=i.sheet_number))
        elif isinstance(i, tuple):
            issue = Issue(itype=IType(i[0]), description=i[1],
                          location=IssueLocation(sheet_name=""))
        else:  # isinstance(i, Issue):
            issue = i

        if issue.itype == IType.ERROR:
            errors_exist = True

        if not issue.ctype and cmd:  # "cmd" may be "None", in case the Issue is produced by the commands container loop
            issue.ctype = cmd._serialization_type

        if not issue.location.sheet_name or issue.location.sheet_name == "":
            issue.location.sheet_name = cmd._source_block_name if hasattr(cmd, "_source_block_name") else ""

        if not issue.location.sheet_number:
            issue.location.sheet_number = sheet_number

        new_issues.append(issue)

    return new_issues, errors_exist


def convert_generator_to_native(generator_type, file_type: str, file):
    """
    Converts a generator
    Creates a generator parser, then it feeds the file type and the file
    The generator parser has to parse the file and to elaborate a native generator (JSON)

    :param generator_type:
    :param file_type:
    :param file:
    :return: Issues and output file
    """

    output = []
    if generator_type.lower() not in ["json", "native", "primitive"]:
        # Create commands generator from factory (from generator_type and file_type)
        state = State()
        commands_generator = commands_container_parser_factory(generator_type, file_type, file, state)
        # Loop over the IExecutableCommand instances
        for cmd, issues in commands_generator:
            # If there are syntax ERRORS, STOP!!!
            stop = False
            if issues and len(issues) > 0:
                for t in issues:
                    if t["type"] == 3:  # Error
                        stop = True
                        break

            output.append({"command": cmd._serialization_type,
                           "label": cmd._serialization_label,
                           "content": cmd.json_serialize(),
                           "issues": issues
                           }
                          )
            if stop:
                break

    return output


# ######################################################################################################################
# SOLVING (PREPARATION AND CALL SOLVER)
# ######################################################################################################################

def prepare_and_solve_model(state: State, dynamic_scenario_parameters: Dict = None) -> List[Issue]:
    """
    Modify the state so that:
    * Implicit references of Interfaces to subcontexts are materialized
      * Creating processors
      * Creating interfaces in these processors
      * Creating relationships in these processors

    * The ProblemStatement class is considered for solving
    q* State is modified to contain the scalar and matrix indicators

    :param state:
    :param dynamic_scenario_parameters:
    :return:
    """
    prepare_model(state)
    issues = call_solver(state, dynamic_scenario_parameters)

    return issues


def call_solver(state: State, dynamic_scenario_parameters: Dict) -> List[Issue]:
    """
    Solve the problem

    :param state: MuSIASEM object model
    :param systems:
    :param dynamic_scenario_parameters: A dictionary containing a dynamic scenario, for interactive exploration
    """

    def obtain_problem_statement(dynamic_scenario_parameters: Dict = None) -> ProblemStatement:
        """
        Obtain a ProblemStatement instance
        Obtain the solver parameters plus a list of scenarios
        :param dynamic_scenario_parameters:
        :return:
        """
        if dynamic_scenario_parameters is not None:
            scenarios = create_dictionary()
            scenarios["dynamic"] = create_dictionary(dynamic_scenario_parameters)
            return ProblemStatement(scenarios=scenarios)
        else:
            ps_list: List[ProblemStatement] = glb_idx.get(ProblemStatement.partial_key())
            if len(ps_list) == 0:
                # No scenarios (dummy), and use the default solver
                scenarios = create_dictionary()
                scenarios["default"] = create_dictionary()
                return ProblemStatement(scenarios=scenarios)
            else:
                # TODO Combine all ProblemStatements into a single ProblemStatement
                return ps_list[-1]  # Take last ProblemStatement

    # Registry and the other objects also
    glb_idx, _, _, datasets, _ = get_case_study_registry_objects(state)

    global_parameters: List[Parameter] = glb_idx.get(Parameter.partial_key())

    dynamic_scenario = dynamic_scenario_parameters is not None
    if not dynamic_scenario:
        problem_statement = obtain_problem_statement()
    else:
        problem_statement = obtain_problem_statement(dynamic_scenario_parameters)

    # Obtain "parameters" Dataset
    datasets["params"] = obtain_parameters_dataset(global_parameters, problem_statement)

    solver_type_param = glb_idx.get(Parameter.partial_key("NISSolverType"))
    solver_type_param = solver_type_param[0]
    solver_type = solver_type_param.current_value

    issues: List[Issue] = []
    if solver_type == "FlowGraph":
        issues = flow_graph_solver(global_parameters, problem_statement, state, dynamic_scenario)

    return issues


def prepare_model(state) -> NoReturn:
    """
    Modify the state so that:
    * Implicit references of Interfaces to subcontexts are materialized
      * Creating processors
      * Creating interfaces in these processors
      * Creating relationships in these processors

    :param state:
    """

    # TODO: currently when an interface is defined as a Scale from two or more interfaces, the computed values are
    #  added while the intuition tells us that only one scale should be defined. We have to give a warning message
    #  if this situation happens.

    # Registry and the other objects also
    glb_idx, _, _, _, _ = get_case_study_registry_objects(state)
    # Prepare a Query to obtain ALL interfaces
    query = BasicQuery(state)
    filt = {}
    objs = query.execute([Factor], filt)
    for iface in objs[Factor]:  # type: Factor
        if strcmp(iface.processor.instance_or_archetype, 'Archetype') or strcmp(iface.processor.instance_or_archetype, 'No'):
            continue

        # If the Interface is connected to a "Subcontext" different than the owning Processor
        if iface.opposite_processor_type:
            if iface.opposite_processor_type.lower() != iface.processor.subsystem_type.lower():
                # Check if the interface has flow relationships
                # TODO An alternative is to search "observations" of type FactorsRelationDirectedFlowObservation
                #      in the same "iface"

                if iface.orientation.lower() == "input":
                    parameter = {"target": iface}
                else:
                    parameter = {"source": iface}

                relations = glb_idx.get(FactorsRelationDirectedFlowObservation.partial_key(**parameter))

                # If it does not have flow relationships:
                #  * define default Processor name and retrieve it (or if it does not exist, create it)
                #  * create an Interface into that Processor and a Flow Relationship
                if len(relations) == 0:
                    # Define the name of a Processor in the same context but in different subcontext
                    p_name = iface.opposite_processor_type + "_" + iface.processor.processor_system
                    p = glb_idx.get(Processor.partial_key(p_name))
                    if len(p) == 0:
                        attributes = {
                            'subsystem_type': iface.opposite_processor_type,
                            'processor_system': iface.processor.processor_system,
                            'functional_or_structural': 'Functional',
                            'instance_or_archetype': 'Instance'
                            # 'stock': None
                        }

                        p = Processor(p_name, attributes=attributes)
                        glb_idx.put(p.key(), p)
                    else:
                        p = p[0]

                    attributes = {
                        'sphere': 'Technosphere' if iface.opposite_processor_type.lower() in ["local", "external"] else 'Biosphere',
                        'roegen_type': iface.roegen_type,
                        'orientation': "Input" if iface.orientation.lower() == "output" else "Output",
                        'opposite_processor_type': iface.processor.subsystem_type
                    }

                    # Create Interface (if it does not exist)
                    if not p.factors_find(iface.taxon.name):
                        f = Factor.create_and_append(name=iface.taxon.name,
                                                     processor=p,
                                                     in_processor_type=
                                                     FactorInProcessorType(external=False,
                                                                           incoming=iface.orientation.lower() == "output"),
                                                     attributes=attributes,
                                                     taxon=iface.taxon)

                        glb_idx.put(f.key(), f)

                    # Create Flow Relationship
                    if iface.orientation.lower() == "output":
                        source = iface
                        target = f
                    else:
                        source = f
                        target = iface

                    fr = FactorsRelationDirectedFlowObservation.create_and_append(
                        source=source,
                        target=target,
                        observer=None)
                    glb_idx.put(fr.key(), fr)


def obtain_parameters_dataset(global_parameters: List[Parameter], problem_statement: ProblemStatement):
    params_keys = []
    params_data = []
    for scenario_name, scenario_exp_params in problem_statement.scenarios.items():  # type: str, dict
        p = evaluate_parameters_for_scenario(global_parameters, scenario_exp_params)
        for k, v in p.items():
            params_keys.append((scenario_name, k))
            params_data.append(v)

    df = pd.DataFrame(params_data,
                      index=pd.MultiIndex.from_tuples(params_keys, names=["Scenario", "Parameter"]),
                      columns=["Value"])
    return get_dataset(df, "params", "Parameter values per Scenario")

# #####################################################################################################################
# >>>> INTERACTIVE SESSION <<<<
# #####################################################################################################################


class CreateNew(Enum):
    CASE_STUDY = 1
    VERSION = 2
    NO = 3


class InteractiveSession:
    """ 
    Main class for interaction with NIS
    The first thing would be to identify the user and create a GUID for the session which can be used by the web server
    to store and retrieve the interactive session state.
    
    It receives command_executors, modifying state accordingly
    If a reproducible session is opened, 
    """
    def __init__(self, session_factory):
        # Session factory with access to business logic database
        self._session_factory = session_factory

        # Interactive session ID
        self._guid = str(uuid.uuid4())

        # User identity, if given (can be an anonymous session)
        self._identity = None  # type: Identity
        self._state = State()  # To keep the state
        self._reproducible_session = None  # type: ReproducibleSession

    def reset_state(self):
        """ Restart state """
        self._state = State()
        # TODO self._recordable_session = None ??

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, state: State):
        self._state = state

    def get_sf(self):
        return self._session_factory

    def set_sf(self, session_factory):
        self._session_factory = session_factory
        if self._reproducible_session:
            self._reproducible_session.set_sf(session_factory)

    def open_db_session(self):
        return self._session_factory()

    def close_db_session(self):
        self._session_factory.remove()

    def quit(self):
        """
        End interactive session
        :return: 
        """
        self.close_reproducible_session()
        self.close_db_session()

    # --------------------------------------------------------------------------------------------

    def identify(self, identity_information, testing=False):
        """
        Given credentials of some type -identity_information-, link an interactive session to an identity.
        The credentials can vary from an OAuth2 Token to user+password.
        Depending on the type of credentials, invoke a type of "identificator" or other
        An interactive session without identification is allowed to perform a subset of available operations
        
        :param identity_information: 
        :return: True if the identification was successful, False if not 
        """
        # TODO Check the credentials
        if isinstance(identity_information, dict):
            if "user" in identity_information and testing:
                # Check if the user is in the User's table
                session = self._session_factory()
                src = session.query(User).filter(User.name == identity_information["user"]).first()
                # Check if the dataset exists. "ETL" it if not
                # ds = session.query(Dataset).\
                #     filter(Dataset.code == dataset).\
                #     join(Dataset.database).join(Database.data_source).\
                #     filter(DataSource.name == src_name).first()
                force_load(src)
                session.close()
                self._session_factory.remove()
                if src:
                    self._identity = src.name
                    if self._state:
                        self._state.set("_identity", self._identity)

                return src is not None
            elif "token" in identity_information:
                # TODO Validate against some Authentication service
                pass

    def get_identity_id(self):
        return self._identity

    def unidentify(self):
        # TODO The un-identification cannot be done in the following circumstances: any?
        self._identity = None
        if self._state:
            self._state.set("_identity", self._identity)

    # --------------------------------------------------------------------------------------------
    # Reproducible sessions and commands INSIDE them
    # --------------------------------------------------------------------------------------------
    def open_reproducible_session(self,
                                  case_study_version_uuid: str,
                                  recover_previous_state=True,
                                  cr_new: CreateNew = CreateNew.NO,
                                  allow_saving=True):
        self._reproducible_session = ReproducibleSession(self)
        self._reproducible_session.open(self._session_factory, case_study_version_uuid, recover_previous_state, cr_new, allow_saving)

    def close_reproducible_session(self, issues=None, output=None, save=False, from_web_service=False, cs_uuid=None, cs_name=None):
        if self._reproducible_session:
            if save:
                # TODO Save issues AND (maybe) output
                self._reproducible_session.save(from_web_service, cs_uuid=cs_uuid, cs_name=cs_name)
            uuid_, v_uuid, cs_uuid = self._reproducible_session.close()
            self._reproducible_session = None
            return uuid_, v_uuid, cs_uuid
        else:
            return None, None, None

    def reproducible_session_opened(self):
        return self._reproducible_session is not None

    @property
    def reproducible_session(self):
        return self._reproducible_session

    # --------------------------------------------------------------

    def execute_executable_command(self, cmd: IExecutableCommand):
        return execute_command(self._state, cmd)

    def register_executable_command(self, cmd: IExecutableCommand):
        self._reproducible_session.register_executable_command(cmd)

    def register_andor_execute_command_generator1(self, c: CommandsContainer, register=True, execute=False, ignore_imports=False):
        """
        Creates a generator parser, then it feeds the file type and the file
        The generator parser has to parse the file and to generate command_executors as a Python generator

        :param generator_type:
        :param file_type:
        :param file:
        :param register: If True, register the command in the ReproducibleSession
        :param execute: If True, execute the command in the ReproducibleSession
        :return:
        """
        if not self._reproducible_session:
            raise Exception("In order to execute a command generator, a work session is needed")
        if not register and not execute:
            raise Exception("More than zero of the parameters 'register' and 'execute' must be True")

        if register:
            self._reproducible_session.register_persistable_command(c)

        if execute:
            c.execution_start = datetime.datetime.now()
            pass_case_study = self._reproducible_session._session.version.case_study is not None
            ret = self._reproducible_session.execute_command_generator(c, pass_case_study, ignore_imports)
            c.execution_end = datetime.datetime.now()
            return ret
            # Or
            # return execute_command_container(self._state, c)
        else:
            return None

    def register_andor_execute_command_generator(self, generator_type, file_type: str, file, register=True, execute=False, ignore_imports=False):
        """
        Creates a generator parser, then it feeds the file type and the file
        The generator parser has to parse the file and to generate command_executors as a Python generator

        :param generator_type: 
        :param file_type: 
        :param file: 
        :param register: If True, register the command in the ReproducibleSession
        :param execute: If True, execute the command in the ReproducibleSession
        :return: 
        """

        return self.register_andor_execute_command_generator1(
            CommandsContainer.create(generator_type, file_type, file),
            register,
            execute,
            ignore_imports
        )

    # --------------------------------------------------------------------------------------------

    def get_case_studies(self):
        """ Get a list of case studies READABLE by current identity (or public if anonymous) """
        pass

    def get_case_study_versions(self, case_study: str):
        # TODO Check that the current user has READ access to the case study
        pass

    def get_case_study_version(self, case_study_version: str):
        # TODO Check that the current user has READ access to the case study
        pass

    def get_case_study_version_variables(self, case_study_version: str):
        """ A tree of variables, by type: processors, flows """
        pass

    def get_case_study_version_variable(self, case_study_version: str, variable: str):
        pass

    def remove_case_study_version(self, case_study_version: str):
        pass

    def share_case_study(self, case_study: str, identities: List[str], permission: str):
        pass

    def remove_case_study_share(self, case_study: str, identities: List[str], permission: str):
        pass

    def get_case_study_permissions(self, case_study: str):
        pass

    def export_case_study_version(self, case_study_version: str):
        pass

    def import_case_study(self, file):
        pass

# #####################################################################################################################
# >>>> REPRODUCIBLE SESSION <<<<
# #####################################################################################################################


class ReproducibleSession:
    def __init__(self, isess):
        # Containing InteractiveSession. Used to set State when a ReproducibleSession is opened and it overwrites State
        self._isess = isess  # type: InteractiveSession
        self._identity = isess._identity
        self._sess_factory = None
        self._allow_saving = None
        self._session = None  # type: CaseStudyVersionSession

    @property
    def ws_commands(self):
        return self._session.commands if self._session else None

    def open(self, session_factory, uuid_: str=None, recover_previous_state=True, cr_new:CreateNew=CreateNew.NO, allow_saving=True):
        """
        Open a work session

    +--------+--------+---------+-----------------------------------------------------------------------------------+
    | UUID   | cr_new | recover |        Behavior                                                                   |
    +--------+--------+---------+-----------------------------------------------------------------------------------+ 
    | !=None | True   | True    | Create new CS or version (branch) from "UUID", clone WS, recover State, append WS |
    | !=None | True   | False   | Create new CS or version (branch) from "UUID", Zero State, first WS               |
    | !=None | False  | True    | Recover State, append WS                                                          |
    | !=None | False  | False   | Zero State, append WS (overwrite type)                                            |
    | ==None | -      | -       | New CS and version, Zero State, first WS                                          |
    +--------+--------+---------+-----------------------------------------------------------------------------------+
    Use cases:
    * A new case study, from scratch. uuid_=None
    * A new case study, copied from another case study. uuid_=<something>, cr_new=CreateNew.CASE_STUDY, recover_previous_state=True
    * A new version of a case study
      - Copying previous version
      - Starting from scratch
    * Continue a case study version
      - But restart (from scratch)
    * Can be a Transient session

        :param uuid_: UUID of the case study or case study version. Can be None, for new case studies or for testing purposes.
        :param recover_previous_state: If an existing version is specified, it will recover its state after execution of all command_executors
        :param cr_new: If != CreateNew.NO, create either a case study or a new version. If == CreateNew.NO, append session to "uuid"
        :param allow_saving: If True, it will allow saving at the end (it will be optional). If False, trying to save will generate an Exception
        :return UUID of the case study version in use. If it is a new case study and it has not been saved, the value will be "None"
        """

        # TODO Just register for now. But in the future it should control that there is no other "allow_saving" ReproducibleSession opened
        # TODO for the same Case Study Version. So it implies modifying some state in CaseStudyVersion to have the UUID
        # TODO of the active ReproducibleSession, even if it is not in the database. Register also the date of "lock", so the
        # TODO lock can be removed in case of "hang" of the locker ReproducibleSession
        self._allow_saving = allow_saving
        self._sess_factory = session_factory
        session = self._sess_factory()
        if uuid_:
            uuid_ = str(uuid_)
            # Find UUID. Is it a Case Study or a Case Study version?
            # If it is the former, look for the active version.
            cs = session.query(CaseStudy).filter(CaseStudy.uuid == uuid_).first()
            if not cs:
                vs = session.query(CaseStudyVersion).filter(CaseStudyVersion.uuid == uuid_).first()
                if not vs:
                    ss = session.query(CaseStudyVersionSession).filter(CaseStudyVersionSession.uuid == uuid_).first()
                    if not ss:
                        raise Exception("Object '"+uuid_+"' not found, when opening a ReproducibleSession")
                    else:
                        vs = ss.version
                        cs = vs.case_study
                else:
                    cs = vs.case_study
            else:  # A case study, find the latest version (the version modified latest -by activity, newest ReproducibleSession-)
                max_date = None
                max_version = None
                for v in cs.versions:
                    for s in v.sessions:
                        if not max_date or s.open_instant > max_date:
                            max_date = s.open_instant
                            max_version = v
                vs = max_version
                cs = vs.case_study

            # List of active sessions
            # NOTE: instead of time ordering, the ID is used, assuming sessions with greater ID were created later
            lst = session.query(CaseStudyVersionSession). \
                filter(CaseStudyVersionSession.version_id == vs.id). \
                order_by(CaseStudyVersionSession.id). \
                all()
            idx = 0
            for i, ws in enumerate(lst):
                if ws.restarts:
                    idx = i
            lst = lst[idx:]  # Cut the list, keep only active sessions

            if cr_new != CreateNew.NO:  # Create either a case study or a case study version
                if cr_new == CreateNew.CASE_STUDY:
                    cs = copy.copy(cs)  # New Case Study: COPY CaseStudy
                else:
                    force_load(cs)  # New Case Study Version: LOAD CaseStudy (then version it)
                vs2 = copy.copy(vs)  # COPY CaseStudyVersion
                vs2.case_study = cs  # Assign case study to the new version
                if recover_previous_state:  # If the new version keeps previous state, copy it also
                    vs2.state = vs.state  # Copy state
                    vs2.state_version = vs.state_version
                    for ws in lst:  # COPY active ReproducibleSessions
                        ws2 = copy.copy(ws)
                        ws2.version = vs2
                        for c in ws.commands:  # COPY commands
                            c2 = copy.copy(c)
                            c2.session = ws2
                vs = vs2
            else:
                # Load into memory
                if len(lst) == 1:
                    ws = lst[0]
                    force_load(ws)
                force_load(vs)
                force_load(cs)

            if recover_previous_state:
                # Load state if it is persisted (if not EXECUTE, POTENTIALLY VERY SLOW)
                if vs.state:
                    # Deserialize
                    self._isess._state = deserialize_state(vs.state, vs.state_version)
                else:
                    self._isess._state = State()  # Zero State, execute all commands in sequence
                    for ws in lst:
                        for c in ws.commands:
                            execute_command_container(self._isess._state, c)
                if cr_new == CreateNew.VERSION:  # TODO Check if this works in all possible circumstances (combine the parameters of the function)
                    recover_previous_state = False
            else:
                self._isess._state = State()

        else:  # New Case Study AND new Case Study Version
            cs = CaseStudy()
            vs = CaseStudyVersion()
            vs.creation_instant = datetime.datetime.utcnow()
            vs.case_study = cs

        # Detach Case Study and Case Study Version
        if cs in session:
            session.expunge(cs)
        if vs in session:
            session.expunge(vs)
        # Create the Case Study Version Session
        usr = session.query(User).filter(User.name == self._identity).first()
        if usr:
            force_load(usr)
        else:
            if allow_saving:
                raise Exception("A user is required to register which user is authoring a case study")
        # TODO !!!!NEW CODE, ADDED TO SUPPORT NEEDED FUNCTIONALITY. NEEDS BETTER CODING!!!!
        restart = not recover_previous_state if uuid_ else True
        if not restart:
            self._session = ws
        else:
            self._session = CaseStudyVersionSession()
            self._session.version = vs
            self._session.who = usr
            self._session.restarts = True
        # If the Version existed, define "restarts" according to parameter "recover_previous_state"
        # ElseIf it is the first Session -> RESTARTS=True

        session.close()
        # session.expunge_all()
        self._sess_factory.remove()

    def update_current_version_state(self, lst_cmds):
        """ Designed to work using the REST interface. TEST in direct use. """
        # Version
        # v = self._session.version
        # Serialize state
        st = serialize_state(self._isess._state)
        # v.state = st
        # Open DB session
        session = self._sess_factory()
        # Load version and change its state
        v = session.query(CaseStudyVersion).get(self._session.version_id)
        v.state = st
        session.add(v)
        for c in lst_cmds:
            c2 = session.query(CommandsContainer).get(c.id)
            c2.execution_start = c.execution_start
            c2.execution_end = c.execution_end
            session.add(c2)
        session.commit()
        self._sess_factory.remove()

    def save(self, from_web_service=False, cs_uuid=None, cs_name=None):
        if not self._allow_saving:
            raise Exception("The ReproducibleSession was opened disallowing saving. Please close it and reopen it with the proper value")
        # Serialize state
        st = serialize_state(self._isess._state)
        self._session.version.state = st
        self._session.state = st
        ws = self._session

        # Open DB session
        session = self._sess_factory()
        # Change the case study
        if cs_uuid:
            # Load case study
            cs = session.query(CaseStudy).filter(CaseStudy.uuid == cs_uuid).first()
            if cs:
                ws.version.case_study = cs
            else:
                raise Exception("The case study UUID '"+cs_uuid+"' was not found")
        # Append commands, self._session, the version and the case_study
        if not from_web_service:
            for c in self._session.commands:
                session.add(c)
            session.add(ws)
            session.add(ws.version)
            session.add(ws.version.case_study)
        else:
            ws.who = session.merge(ws.who)
            cs_id = ws.version.case_study.id
            vs_id = ws.version.id
            if cs_id and not vs_id:
                ws.version.case_study = None
            if vs_id:
                ws.version = None

            if cs_id:
                cs = session.query(CaseStudy).get(cs_id)
            else:
                cs = ws.version.case_study
                session.add(cs)

            if vs_id:
                vs = session.query(CaseStudyVersion).get(vs_id)
                ws.version = vs
            else:
                ws.version.case_study = cs
                vs = ws.version
                session.add(vs)

            ws.close_instant = datetime.datetime.utcnow()
            session.add(ws)
            for c in self._session.commands:
                session.add(c)
        if cs_name:
            ws.version.name = cs_name

        # If it was called from the REST API, assure that the version has a creation date (it should not happen)
        if from_web_service and not vs.creation_instant:
            logging.debug("Late setup of version creation date")
            vs.creation_instant = datetime.datetime.utcnow()

        # Commit DB session
        session.commit()
        force_load(self._session)
        self._sess_factory.remove()

    def register_persistable_command(self, cmd: CommandsContainer):
        cmd.session = self._session

    def create_and_register_persistable_command(self, generator_type, file_type, file):
        """
        Generates command_executors from an input stream (string or file)
        There must be a factory to parse stream 
        :param generator_type: 
        :param file_type: 
        :param file: It can be a stream or a URL or a file name
        """
        c = CommandsContainer.create(generator_type, file_type, file)
        self.register_persistable_command(c)
        return c

    def execute_command_generator(self, cmd: CommandsContainer, pass_case_study=False, ignore_imports=False):
        if pass_case_study:  # CaseStudy can be modified by Metadata command, pass a reference to it
            self._isess._state.set("_case_study", self._session.version.case_study)
            self._isess._state.set("_case_study_version", self._session.version)

        ret = execute_command_container(self._isess._state, cmd, ignore_imports)

        if pass_case_study:
            self._isess._state.set("_case_study", None)
            self._isess._state.set("_case_study_version", None)

        return ret

    def register_executable_command(self, command: IExecutableCommand):
        c = executable_command_to_commands_container(command)
        c.session = self._session

    def set_sf(self, session_factory):
        self._sess_factory = session_factory

    @property
    def commands(self):
        return self._session.commands

    @property
    def case_study(self):
        return self._session.version.case_study

    def close(self) -> tuple:
        if not self._session:
            raise Exception("The CaseStudyVersionSession is not opened")
        id3 = self._session.uuid, self._session.version.uuid, self._session.version.case_study.uuid
        self._session = None
        self._allow_saving = None
        return id3


def execute_file_return_issues(file_name, generator_type):
    """
    Execution of files in the context of TESTS

    :param file_name:
    :param generator_type:
    :return:
    """
    if generator_type == "spreadsheet":
        content_type = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
        read_type = "rb"
    elif generator_type == "native":
        content_type = "application/json"
        read_type = "r"

    prepare_and_reset_database_for_tests()
    isess = InteractiveSession(DBSession)
    isess.identify({"user": "test_user"}, testing=True)  # Pass just user name.
    isess.open_reproducible_session(case_study_version_uuid=None,
                                    recover_previous_state=None,
                                    cr_new=CreateNew.CASE_STUDY,
                                    allow_saving=False)

    # Add system-level entities from JSON definition in "default_cmds"
    ret = isess.register_andor_execute_command_generator("json", "application/json", nexinfosys.default_cmds, False, True)

    # Execute current file
    with open(file_name, read_type) as f1:
        buffer = f1.read()

    issues, output = isess.register_andor_execute_command_generator(generator_type, content_type, buffer, False, True)

    for idx, issue in enumerate(issues):
        logging.debug(f"Issue {idx+1}/{len(issues)} = {issue}")

    logging.debug(f"Output = {output}")

    isess.close_reproducible_session()
    isess.close_db_session()
    return isess, issues


def execute_file(file_name, generator_type):
    """
    Execution of files in the context of TESTS

    :param file_name:
    :param generator_type:
    :return:
    """
    return execute_file_return_issues(file_name, generator_type)[0]  # Return just "isession"


if __name__ == '__main__':
    import jsonpickle
    with open("/home/rnebot/pickled_state", "r") as f:
        s = f.read()
    o = jsonpickle.decode(s)
    # Submit Worksheet as New Case Study
    isess = InteractiveSession()
    isess.quit()




"""
COMEXT database

This database is Eurostat, but uses an organization differentiated from the rest of datasets.

An extract of mail messages between Eurostat's support and MAGIC development is at the end of the file (Spanish)

"""

import os
import re
import tempfile
from io import StringIO, BytesIO
from abc import abstractmethod
import zipfile
from typing import List
import pandas as pd
import numpy as np
import requests

from nexinfosys.common.helper import translate_case
from nexinfosys.ie_imports.data_source_manager import IDataSourceManager, get_dataset_structure, \
    filter_dataset_into_dataframe
from nexinfosys.ie_imports.data_sources.eurostat_bulk import Eurostat
from nexinfosys.models.statistical_datasets import DataSource, Database, Dataset, Dimension


dataset_csv = """
Code,File,Label
DS-066341,epannsold-r2.zip,"Sold production, exports and imports by PRODCOM list (EUROPROMS Nace r2)"
DS-066342,epanntotal-r2.zip,"Total production by PRODCOM list (EUROPROMS NACE r2)"
DS-043408,epannsold.zip,"Sold production, exports and imports by PRODCOM list (EUROPROMS Nace r1)"
DS-043409,epanntotal.zip,"Total production by PRODCOM list (EUROPROMS NACE r1)"
"""

# Map of correspondence of field names, DS-066341 -> DS-043408
# Also valid for correspondence of field names between DS-066342 and DS-043409
map_r2_r1_fields = {
    "DECL": "DECLARANT",
    "PRCCODE": "PRCCODE",
    "PERIOD": "PERIOD",
    "EXPQNT": "EXP_QUANTITY",
    "EXPVAL": "EXP_VALUE",
    "IMPQNT": "IMP_QUANTITY",
    "IMPVAL": "IMP_VALUE",
    "PRODQNT": "PROD_QUANTITY",
    "PVALBASE": "PROD_VALUE_BASE",
    "PQNTFLAG": "PROD_QUANTITY_FLAG",
    "PQNTBASE": "PROD_QUANTITY_BASE",
    "PRODVAL": "PROD_VALUE_EUR",
    "PVALFLAG": "PROD_VALUE_FLAG",
    "QNTUNIT": "UNIT"
}

dimension_descriptions = {
    "DECL": "Country",
    "PRCCODE": "NACE code",
    "PERIOD": "Year",
    "EXPQNT": "Exported quantity",
    "EXPVAL": "Exported price",
    "IMPQNT": "Imported quantity",
    "IMPVAL": "Imported price",
    "PRODQNT": "Sold production quantity",
    "PRODVAL": "Sold production price",
    "PVALBASE": "Sold production price BASE",
    "PVALFLAG": "Sold production price FLAG",
    "PQNTFLAG": "Sold production quantity FLAG",
    "PQNTBASE": "Sold production quantity BASE",
    "QNTUNIT": "Unit for quantities"
}


class COMEXT(IDataSourceManager):
    def __init__(self):
        self._eurostat = Eurostat()  # Used to obtain datasets Metadata
        self._bulk_download_url = "https://ec.europa.eu/eurostat/estat-navtree-portlet-prod/BulkDownloadListing?" \
                                  "sort=1&file=comext%2FCOMEXT_OTHER_DATA%2FEUROPROMS%2F"
        self._datasets = pd.read_csv(StringIO(dataset_csv))
        self._datasets["idx"] = self._datasets["Code"]
        self._datasets.set_index("idx", inplace=True)

    def get_name(self) -> str:
        """ Source name """
        return self.get_datasource().name

    def get_datasource(self) -> DataSource:
        """ Data source """
        src = DataSource()
        src.name = "COMEXT"
        src.description = "Eurostat's COMEXT"
        return src

    def get_databases(self) -> List[Database]:
        """ List of databases in the data source """
        db = Database()
        db.code = ""
        db.description = "Comext"
        return [db]

    def get_datasets(self, database=None) -> list:
        """ List of datasets in a database, or in all the datasource (if database==None)
            Return a list of tuples (database, dataset)
        """
        return [(r["Code"], r["Label"], r["Code"]) for i, r in self._datasets.iterrows()]

    def get_dataset_structure(self, database, dataset) -> Dataset:
        """ Obtain the structure of a dataset: concepts, dimensions, attributes and measures """
        ds_tmp = self._eurostat.get_dataset_structure(None, dataset)
        dims = {d.code: d for d in ds_tmp.dimensions}

        ds = Dataset()
        ds.code = ds_tmp.code
        ds.description = ds_tmp.description
        ds.database = ds_tmp.database

        # Dimensions are: DECLARANT (Country) and PRCCODE (NACE + others Code), PERIOD (Time)
        declarant = dims["DECL" if dataset.startswith("DS-066") else map_r2_r1_fields["DECL"]]
        dd = Dimension()
        dd.code = "DECL" if dataset.startswith("DS-066") else map_r2_r1_fields["DECL"]
        dd.description = dimension_descriptions["DECL"]
        dd.attributes = None
        dd.is_time = None
        dd.is_measure = None
        dd.dataset = ds
        dd.code_list = declarant.code_list.clone()
        # dd.code_list.dimension = dd

        prccode = dims["PRCCODE"]
        dd = Dimension()
        dd.code = "PRCCODE" if dataset.startswith("DS-066") else map_r2_r1_fields["PRCCODE"]
        dd.description = dimension_descriptions["PRCCODE"]
        dd.attributes = None
        dd.is_time = None
        dd.is_measure = None
        dd.dataset = ds
        dd.code_list = prccode.code_list.clone()
        # dd.code_list.dimension = dd

        dd = Dimension()
        dd.code = "PERIOD" if dataset.startswith("DS-066") else map_r2_r1_fields["PERIOD"]
        dd.description = dimension_descriptions["PERIOD"]
        dd.attributes = None
        dd.is_time = True
        dd.is_measure = None
        dd.dataset = ds

        if dataset in ("DS-066341", "DS-043408"):
            measures = ["PRODQNT", "EXPQNT", "IMPQNT", "PQNTBASE", "PRODVAL", "EXPVAL", "IMPVAL", "PVALBASE"]
            attributes = ["QNTUNIT", "PVALFLAG", "PQNTFLAG"]
        else:
            measures = ["PRODQNT", "PQNTBASE"]
            attributes = ["PQNTFLAG", "QNTUNIT"]

        for m in measures:
            dd = Dimension()
            dd.code = m if dataset.startswith("DS-066") else map_r2_r1_fields[m]
            dd.description = dimension_descriptions[m]
            dd.attributes = None
            dd.is_time = None
            dd.is_measure = True
            dd.dataset = ds
        # TODO Attributes

        return ds

    def etl_full_database(self, database=None, update=False):
        """ If bulk download is supported, refresh full database """
        pass

    def etl_dataset(self, dataset, update=False):
        """ If bulk download is supported, refresh full dataset """
        file_base = self._datasets.loc[dataset, "File"]
        url = self._bulk_download_url + file_base
        zip_name = tempfile.gettempdir() + "/" + file_base
        if os.path.isfile(zip_name):
            if not update:
                return zip_name
            else:
                os.remove(zip_name)

        r = requests.get(url, stream=True)
        # http://stackoverflow.com/questions/15352668/download-and-decompress-gzipped-file-in-memory
        with open(zip_name, 'wb') as f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)

        return zip_name

    def get_dataset_filtered(self, dataset: str, dataset_params: list) -> Dataset:
        """ Obtains the dataset with its structure plus the filtered values
            The values can be in a pd.DataFrame or in JSONStat compact format
            After this, new dimensions can be joined, aggregations need to be performed
        """
        def multi_replace(text, rep):
            rep = dict((re.escape(k), v) for k, v in rep.items())
            pattern = re.compile("|".join(rep.keys()))
            return pattern.sub(lambda m: rep[re.escape(m.group(0))], text)

        # Read dataset structure
        ds = self.get_dataset_structure(None, dataset)

        # Read full dataset into a Dataframe
        dataframe_fn = tempfile.gettempdir() + "/" + dataset + ".bin2"
        df = None
        if os.path.isfile(dataframe_fn):
            df = pd.read_parquet(dataframe_fn)
            # df = pd.read_msgpack(dataframe_fn)

        if df is None:
            zip_name = self.etl_dataset(dataset, update=False)
            with zipfile.ZipFile(zip_name, "r") as arch:
                # Read file and
                st = arch.read(arch.filelist[0].filename)
                # pattern = re.compile("(\\:)|( [b-fnpruzscde]+)")
                # st = pattern.sub(lambda m: "NaN" if m.group(0) == ":" else "", gz.read().decode("utf-8"))

                fc = BytesIO(st)
            # os.remove(zip_name)  # Remove, because the dataset is stored as PARQUET dataframe format
            df = pd.read_csv(fc, dtype={"PRCCODE": "str",
                                        "DECL" if dataset.startswith("DS-066") else map_r2_r1_fields["DECL"]: "str"
                                        }
                             )

            # Get the list of dimensions and of measures
            dimensions = []
            measures = []
            for d in ds.dimensions:
                if d.is_measure:
                    measures.append(d.code)
                else:
                    dimensions.append(d.code)

            # Convert measures to numeric
            to_delete = []
            for cn in df.columns:
                if cn in measures:
                    df[cn] = df[cn].astype(np.float)
                elif cn not in dimensions:
                    to_delete.append(cn)
                elif cn == "PERIOD":
                    # Remove the tail "52" (meaning "year")
                    df[cn] = (df[cn]//100).astype("str")

            # Delete unused columns
            for cn in to_delete:
                del df[cn]

            # Set index on the dimension columns
            df.set_index(dimensions, inplace=True)
            # Save df
            df.to_parquet(dataframe_fn)
            # df.to_msgpack(dataframe_fn)

        # Change dataframe index names to match the case of the names in the metadata
        # metadata_names_dict = {dim.code.lower(): dim.code for dim in ds.dimensions}
        # dataframe_new_names = [metadata_names_dict.get(name.lower(), name) for name in df.index.names]
        dataframe_new_names = translate_case(df.index.names, [dim.code for dim in ds.dimensions])
        df.index.names = dataframe_new_names

        # Filter it using generic Pandas filtering capabilities
        def obtain_periods_to_filter(filter_dict):
            start = None
            if "StartPeriod" in filter_dict:
                start = filter_dict["StartPeriod"]
                del filter_dict["StartPeriod"]
                if isinstance(start, list): start = start[0]
            if "EndPeriod" in filter_dict:
                endd = filter_dict["EndPeriod"]
                del filter_dict["EndPeriod"]
                if isinstance(endd, list): endd = endd[0]
            else:
                if start:
                    endd = start
            if start:
                # Assume year, convert to integer, generate range, then back to string
                start = int(start)
                endd = int(endd)
                filter_dict["PERIOD"] = [str(a) for a in range(start, endd + 1)]

        if dataset_params:
            obtain_periods_to_filter(dataset_params)
            ds.data = filter_dataset_into_dataframe(df, dataset_params, dataset, eurostat_postprocessing=False)
        else:
            ds.data = df

        return ds

    def get_refresh_policy(self):  # Refresh frequency for list of databases, list of datasets, and dataset
        pass


if __name__ == '__main__':
    e = COMEXT()
    e.get_dataset_filtered("DS-066341", None)


"""
QQQQ
Buenas tardes,

Me gustaría saber si existe la posibilidad de descargar por completo (bulk) desde alguna URL los datsets PRODCOM,
además sin requerir interacción humana.

Ahora mismo más de 6000 de los datasets de Eurostat se pueden descargar perfectamente de esta manera.
Sin embargo no he encontrado los de PRODCOM, y era simplemente saber si es que no es posible o si la URL
de descarga es distinta.

Muchas gracias.
---------------------------------------------------------------------------------------------------------------------
AAAA
Gracias por su interés en la página web de Eurostat, www.ec.europa.eu/eurostat, y por plantearnos su consulta telefónica.

He encontrado la opción de descarga para PRODCOM. Le adjunto a continuación el enlace a esos archivos y la ruta seguida.
El único problema es que los datos son de finales de 2017. Hemos preguntado al equipo responsable y, como le acabo de
comentar por teléfono, nos han comunicado que están trabajando en las actualizaciones y que esperan subir nuevos datos
para finales de la próxima semana, en principio.

Aquí tiene el enlace a los datos de PRODCOM:

http://ec.europa.eu/eurostat/estat-navtree-portlet-prod/BulkDownloadListing?sort=1&dir=comext%2FCOMEXT_OTHER_DATA%2FEUROPROMS

Ruta: Bulk Download Listing > comext > COMEXT OTHER DATA > Europroms > seleccione a continuación el archivo o archivos
que más le interesen.

Esperamos que esta información le sea de utilidad.
---------------------------------------------------------------------------------------------------------------------
QQQQ
La información que me pasas ya me sirve bastante.

Para completar el desarrollo que tengo que hacer, necesitaría saber también cómo consultar los metadatos. Con los demás
datasets de Eurostat existe una URL donde se obtiene una información en formato SDMX. Me pasa algo parecido a lo
anterior, y es que no sé qué URL habría para consultar los datasets de COMEXT. Por ejemplo para los demás datasets de
Eurostat esta URL es:

http://ec.europa.eu/eurostat/SDMX/diss-web/rest/dataflow/ESTAT/all/latest

¿Hay alguna equivalente para COMEXT?
---------------------------------------------------------------------------------------------------------------------
AAAA
Buenos días de nuevo:

Me alegro mucho de que le haya servido. Me comunican que para COMEXT y PRODCOM no disponemos de un archivo en formato
SDMX como los otros que ha encontrado. Los metadatos sí que los tenemos en la página de Eurostat para las tablas de
PRODCOM:

http://ec.europa.eu/eurostat/web/prodcom/data/database

En el símbolo con la M al lado del nombre de la carpeta encontrará la información. Le paso el enlace también:

http://ec.europa.eu/eurostat/cache/metadata/en/prom_esms.htm

Espero que, teniendo la información, pueda manejarla usted mismo para su proyecto. En cualquier caso, recuerde mirar
los otros enlaces la semana que viene, por si hubiera una actualización en este sentido.

"""



import configparser
import importlib
import logging
import traceback
from enum import Enum
import os
import regex as re
from typing import Optional, Any, List, Tuple, Callable, Dict, Union, Type

import pint
from collections import namedtuple
from attr import attrs, attrib

# GLOBAL VARIABLES
cfg_file_env_var = "MAGIC_NIS_SERVICE_CONFIG_FILE"
engine = None

# Database containing OLAP data (cache of Data Cubes)
data_engine = None

# Data source manager
data_source_manager = None  # type: DataSourceManager

# REDIS
redis = None

# Use AdHoc datasets source (datasets in memory as a datasource, for DatasetQry)
enable_adhoc_datasets_source = True

# Case sensitive
case_sensitive = False

# Create units registry
ureg = pint.UnitRegistry()
ureg.define("cubic_meter = m^3 = m3")
ureg.define("square_meter = m^2 = m2")
ureg.define("euro = [] = EUR = Eur = eur = Euro = Euros = €")
ureg.define("dollar = [] = USD = Usd = usd = Dollar = Dollars = $")
ureg.define("capita = []")
ureg.define("dimensionless = []")
ureg.define('fraction = [] = frac')
ureg.define('percent = 1e-2 frac = pct')

# Named tuples
Issue = namedtuple("Issue",
                   "sheet_number sheet_name c_type type message")

SDMXConcept = namedtuple('Concept', 'type name istime description code_list')

# Global Types

IssuesOutputPairType = Tuple[List[Issue], Optional[Any]]
ExecutableCommandIssuesPairType = Tuple[Optional["IExecutableCommand"], List[Issue]]
IssuesLabelContentTripleType = Tuple[List[Issue], Optional[Any], Optional[Dict[str, Any]]]
# Tuple (top, bottom, left, right) representing the rectangular area of the input worksheet where the command is present
AreaTupleType = Tuple[int, int, int, int]

# ##################################
# METADATA special variables

# Simple DC fields not covered:
#  type (controlled),
#  format (controlled),
#  rights (controlled),
#  publisher,
#  contributor,
#  relation
#
# XML Dublin Core: http://www.dublincore.org/documents/dc-xml-guidelines/
# Exhaustive list: http://dublincore.org/documents/dcmi-type-vocabulary/

# Fields: ("<field label in Spreadsheet file>", "<field name in Dublin Core>", Mandatory?, Controlled?, NameInJSON)
metadata_fields = [("Case study name", "title", False, False, "case_study_name"),
                   ("Case study code", "title", True, False, "case_study_code"),
                   ("Title", "title", True, False, "title"),
                   ("Subject, topic and/or keywords", "subject", False, True, "subject_topic_keywords"),
                   ("Description", "description", False, False, "description"),
                   ("Geographical level", "description", True, True, "geographical_level"),
                   ("Dimensions", "subject", True, True, "dimensions"),
                   ("Reference documentation", "source", False, False, "reference_documentation"),
                   ("Authors", "creator", True, False, "authors"),
                   ("Date of elaboration", "date", True, False, "date_of_elaboration"),
                   ("Temporal situation", "coverage", True, False, "temporal_situation"),
                   ("Geographical location", "coverage", True, False, "geographical_situation"),
                   ("DOI", "identifier", False, False, "doi"),
                   ("Language", "language", True, True, "language"),
                   ("Restriction level", None, True, True, "restriction_level"),
                   ("Version", None, True, False, "version")
                   ]

# Regular expression definitions
regex_var_name = "([a-zA-Z][a-zA-Z0-9_-]*)"
regex_hvar_name = "(" + regex_var_name + r"(\." + regex_var_name + ")*)"
regex_cplex_var = "((" + regex_var_name + "::)?" + regex_hvar_name + ")"
regex_optional_alphanumeric = "([ a-zA-Z0-9_-]*)?"  # Whitespace also allowed


# Regular expression for "worksheet name" in version 2
def simple_regex(names: List[str]):
    return r"(" + "|".join(names) + ")" + regex_optional_alphanumeric


# #####################################################################################################################
# >>>> CONFIGURATION FILE AND DIRECTORIES <<<<
# #####################################################################################################################

def prepare_default_configuration(create_directories):
    def default_directories(path, tmp_path):
        return {
            "CASE_STUDIES_DIR": f"{path}/data/cs/",
            "FAO_DATASETS_DIR": f"{path}/data/faostat/",
            "FADN_FILES_LOCATION": f"{path}/data/fadn",
            "CACHE_FILE_LOCATION": f"{tmp_path}/sdmx_datasets_cache",
            "REDIS_HOST_FILESYSTEM_DIR": f"{tmp_path}/sessions",
            "SSP_FILES_DIR": "",
        }

    from appdirs import AppDirs
    app_dirs = AppDirs("nis-backend")

    # Default directories, multi-platform
    data_path = app_dirs.user_data_dir
    cache_path = app_dirs.user_cache_dir
    if create_directories:
        logging.debug(f"Creating directories {data_path} and {cache_path}")
        os.makedirs(data_path, exist_ok=True)
        os.makedirs(cache_path, exist_ok=True)

    # Obtain and create directories
    dirs = default_directories(data_path, cache_path)
    for v in dirs.values():
        if v:
            if create_directories:
                logging.debug(f"Creating directory {v}")
                os.makedirs(v, exist_ok=True)

    # Default configuration
    return f"""{os.linesep.join([f'{k}="{v}"' for k, v in dirs.items()])}
DB_CONNECTION_STRING="sqlite:///{data_path}{os.sep}nis_metadata.db"
DATA_CONNECTION_STRING="sqlite:///{data_path}{os.sep}nis_cached_datasets.db"
SQL_ECHO=True
# Flask Session (server side session)
REDIS_HOST="filesystem:local_session"
TESTING="True"
ENABLE_CYTHON_OPTIMIZATIONS="True"
SELF_SCHEMA=""
FS_TYPE="WebDAV"
FS_SERVER=""
FS_USER=""
FS_PASSWORD=""
# Google Drive API
GAPI_CREDENTIALS_FILE="{data_path}/credentials.json"
GAPI_TOKEN_FILE="{data_path}/token.pickle"    
""", data_path + os.sep + "nis_local.conf"

# Global configuration variables
global_configuration = None


def initialize_configuration():
    try:
        _, file_name = prepare_default_configuration(False)
        logging.debug(f"Default configuration file name: {file_name}")
        if not os.environ.get(cfg_file_env_var):
            logging.debug(f"{cfg_file_env_var} not defined, trying default configuration file location ({file_name}")
            found = False
            for f in [file_name]:  # f"{nexinfosys.__path__[0]}/restful_service/nis_local_dist.conf"
                if os.path.isfile(f):
                    logging.debug(
                        f"Default configuration file {file_name} found, setting {cfg_file_env_var} to that file")
                    found = True
                    os.environ[cfg_file_env_var] = f
                    break
            if not found:
                logging.debug(f"Default configuration file {file_name} NOT found, will try to generate a default in that location.")
                cfg, file_name = prepare_default_configuration(True)
                logging.debug(f"Generating {file_name} as configuration file with content:\n{cfg}")
                with open(file_name, "wt") as f:
                    f.write(cfg)
                logging.debug(
                    f"Default configuration file {file_name} generated! Setting {cfg_file_env_var} to that file")
                os.environ[cfg_file_env_var] = file_name
        else:
            logging.debug(f"{cfg_file_env_var}: {os.environ[cfg_file_env_var]}")

    except Exception as e:
        print(f"{cfg_file_env_var} environment variable not defined!")
        print(e)
        traceback.print_exc()


def get_global_configuration():
    def read_configuration() -> Dict[str, str]:
        """
        If environment variable "MAGIC_NIS_SERVICE_CONFIG_FILE" is defined, and the contents is the name of an existing file,
        read it as a configuration file and return the result

        :return:
        """
        if os.environ.get(cfg_file_env_var):
            fname = os.environ[cfg_file_env_var]
            if os.path.exists(fname):
                with open(fname, 'r') as f:
                    config_string = '[asection]\n' + f.read()
                config = configparser.ConfigParser()
                config.read_string(config_string)
                return {k: expand_paths(k, remove_quotes(v)) for k, v in config.items("asection")}
        else:
            return {}

    global global_configuration
    if global_configuration is None:
        global_configuration = read_configuration()
    return global_configuration


def get_global_configuration_variable(key: str, default: str = None) -> str:
    global global_configuration
    get_global_configuration()
    return global_configuration.get(key.lower(), default)


def set_global_configuration_variable(key: str, value: str):
    """
    Use carefully. The initial intention is to set the "ENABLE_CYTHON_OPTIMIZATIONS" variable to False

    :param key:
    :param value:
    :return:
    """
    global global_configuration
    if global_configuration:
        global_configuration[key.lower()] = value


def remove_quotes(s: str) -> str:
    return s.lstrip('\'"').rstrip('\'"')


def expand_paths(key: str, value: str) -> str:
    if key.endswith('_dir') or key.endswith('_location') or key.endswith('_file'):
        return os.path.expanduser(value)
    else:
        return value

# ##################################
# Commands


class CommandType(Enum):
    input = (1, "Input")
    core = (2, "Core")
    convenience = (3, "Convenience")
    metadata = (4, "Metadata")
    analysis = (5, "Analysis")
    misc = (99, "Miscellaneous")

    def __str__(self):
        return self.value[1]

    @classmethod
    def from_string(cls, s):
        for ct in cls:
            if ct.value[1] == s:
                return ct
        raise ValueError(cls.__name__ + ' has no value matching "' + s + '"')


@attrs(cmp=False)  # Constant and Hashable by id
class Command:
    # Name, the lowercase unique name
    name = attrib()  # type: str
    # Allowed names for the worksheet. Used for simple regular expressions.
    allowed_names = attrib()  # type: List[str]
    # Name of the subclass of IExecutableCommand in charge of the execution
    execution_class_name = attrib()  # type: Optional[str]
    # Command type
    cmd_type = attrib()  # type: CommandType
    # Direct examples
    direct_examples = attrib(default=[])  # type: Optional[List[str]]
    # URLs of files where it is used
    files = attrib(default=[])  # type: Optional[List[str]]
    # Alternative regular expression for worksheet name, otherwise the simple_regex() is used
    alt_regex = attrib(default=None)
    # Parse function, having params (Worksheet, Area) and returning a tuple (issues, label, content)
    # Callable[[Worksheet, AreaTupleType, str, ...], IssuesLabelContentTripleType] = attrib(default=None)
    parse_function: Callable[..., IssuesLabelContentTripleType] = attrib(default=None)
    # In which version is this command allowed?
    is_v1 = attrib(default=False)  # type: bool
    is_v2 = attrib(default=False)  # type: bool

    @property
    def regex(self):
        if self.alt_regex:
            pattern = self.alt_regex
        else:
            pattern = simple_regex(self.allowed_names)

        return re.compile("^"+pattern, flags=re.IGNORECASE)

    @property
    def execution_class(self):
        if self.execution_class_name:
            module_name, class_name = self.execution_class_name.rsplit(".", 1)
            return getattr(importlib.import_module(module_name), class_name)
        else:
            return None


@attrs(cmp=False)  # Constant and Hashable by id
class CommandField:
    # Allowed names for the column
    allowed_names = attrib()  # type: List[str]
    # Internal name used during the parsing
    name = attrib()  # type: str
    # Parser for the column
    parser = attrib()
    # Flag indicating if the column is mandatory or optional. It can also be an expression (string).
    mandatory = attrib(default=False)  # type: Union[bool, str]
    # A default value for the field
    default_value = attrib(default=None)
    # Some columns have a predefined set of allowed strings
    allowed_values = attrib(default=None)  # type: Optional[list[str]]
    # Many values or just one
    many_values = attrib(default=True)
    # Many appearances (the field can appear multiple times). A convenience to define a list
    many_appearances = attrib(default=False)
    # Examples
    examples = attrib(default=None)  # type: List[str]
    # Description text
    description = attrib(default=None)  # type: str
    # Compiled regex
    # regex_allowed_names = attrib(default=None)
    # Is it directly an attribute of a Musiasem type? Which one?
    attribute_of = attrib(default=None)  # type: Type
    deprecated = attrib(default=False)  # type: bool

    @property
    def regex_allowed_names(self):
        def contains_any(s, setc):
            return 1 in [c in s for c in setc]

        # Compile the regular expressions of column names
        rep = [(r if contains_any(r, ".+") else re.escape(r))+"$" for r in self.allowed_names]

        return re.compile("|".join(rep), flags=re.IGNORECASE)


# Embedded commands, assumed to be present in every submitted workbook
default_cmds = """
[
    {
      "command": "cat_hierarchies",
      "label": "Hierarchies by default",
      "content":
        {
            "command_name": "Hierarchies",
            "items": [
            {
                "_complex": false,
                "_expandable": [],
                "_row": 2,
                "code": "FlowGraph",
                "description": "Flow graph algorithm for the solving",
                "hierarchy_name": "SolvingAlgorithms",
                "label": "Solve graph converting the MuSIASEM into overlapping flow graphs"
            },
            {
                "_complex": false,
                "_expandable": [],
                "_row": 2,
                "code": "TakeUpper",
                "description": "Policy: take the value upper instead of the value coming from lower level accumulation",
                "hierarchy_name": "AggregationConflictResolutionPolicies",
                "label": "Take value from the upper level"
            },
            {
                "_complex": false,
                "_expandable": [],
                "_row": 3,
                "code": "TakeLowerAggregation",
                "description": "Policy: take the value computed by aggregation lower levels of a hierarchy",
                "hierarchy_name": "AggregationConflictResolutionPolicies",
                "label": "Take value accumulated from lower levels"
            },
            {
                "_complex": false,
                "_expandable": [],
                "_row": 3,
                "code": "UseZero",
                "description": "Policy: use the zero value when a child is missing during hierarchy aggregation",
                "hierarchy_name": "AggregationMissingValueResolutionPolicies",
                "label": "Use zero in aggregations"
            },
            {
                "_complex": false,
                "_expandable": [],
                "_row": 3,
                "code": "Invalidate",
                "description": "Policy: invalidate the aggregation for a node in hierarchy if any of the children's value is missing",
                "hierarchy_name": "AggregationMissingValueResolutionPolicies",
                "label": "Invalidate aggregation"
            }

            ]
        }
    },
    {
      "command": "parameters",
      "label": "Parameters by default",
      "content":
        {
            "command_name": "Parameters",
                    "items": [
                        {
                            "_complex": false,
                            "_expandable": [],
                            "_row": 2,
                            "description": "Which algorithm to use for the solving",
                            "domain": "SolvingAlgorithms",
                            "group": "NISSolverParameters",
                            "name": "NISSolverType",
                            "type": "Code",
                            "value": "FlowGraph"
                        },
                        {
                            "_complex": false,
                            "_expandable": [],
                            "_row": 2,
                            "description": "Ordered list of consideration of observations",
                            "domain": null,
                            "group": "NISSolverParameters",
                            "name": "NISSolverObserversPriority",
                            "type": "String",
                            "value": ""
                        },
                        {
                            "_complex": false,
                            "_expandable": [],
                            "_row": 2,
                            "description": "Conflicting data resolution policy",
                            "domain": "AggregationConflictResolutionPolicies",
                            "group": "NISSolverParameters",
                            "name": "NISSolverAggregationConflictResolutionPolicy",
                            "type": "Code",
                            "value": "TakeUpper"
                        },
                        {
                            "_complex": false,
                            "_expandable": [],
                            "_row": 2,
                            "description": "Missing value resolution policy",
                            "domain": "AggregationMissingValueResolutionPolicies",
                            "group": "NISSolverParameters",
                            "name": "NISSolverMissingValueResolutionPolicy",
                            "type": "Code",
                            "value": "UseZero"
                        },
                        {
                            "_complex": false,
                            "_expandable": [],
                            "_row": 2,
                            "description": "Ordered list of computation sources to consider on conflicts",
                            "domain": null,
                            "group": "NISSolverParameters",
                            "name": "NISSolverComputationSourcesPriority",
                            "type": "String",
                            "value": "Scale, PartOfAggregation, InterfaceTypeAggregation, ScaleChange, Flow"
                        }
                    ]
        }
    }
]
"""


import json
from collections import namedtuple
from typing import List

from attr import attrs, attrib
from sqlalchemy import Column, Integer, String, Unicode, Boolean, ForeignKey
from sqlalchemy import orm, event
from sqlalchemy.orm import relationship, backref

from nexinfosys.common.helper import create_dictionary
from nexinfosys.models import CodeImmutable
from nexinfosys.models.musiasem_concepts import Hierarchy
from nexinfosys.models.musiasem_methodology_support import ORMBase

# ######################################################################################################################
# PERSISTENT STATISTICAL DATASET
# ######################################################################################################################


class CodeList(ORMBase):
    __tablename__ = "dc_cl"

    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(32))
    description = Column(String(1024))

    @orm.reconstructor
    def init_on_load(self):
        pass
        # self.code_list = None

    def get_codes(self):
        return self.to_dict().items()

    def to_dict(self):
        d = {}
        for l in self.levels:
            for c in l.codes:
                d[c.code] = c.description
        return d

    @staticmethod
    def construct_from_dict(d):
        return CodeList.construct(d["name"], d["description"], d["levels"], [CodeImmutable(**i) for i in d["codes"].values])

    @staticmethod
    def construct(name: str, description: str, levels: List[str], codes: List[CodeImmutable]):
        """

        :param name: Name of the Code List
        :param description: Description of the Code List
        :param levels: Names of the levels
        :param codes: List of codes, including in each the following tuple: CodeImmutable = namedtuple("CodeTuple", "code description level children")
        :return:
        """

        cl = CodeList()
        cl.code = name
        cl.description = description
        # Levels
        levels_dict = create_dictionary()
        for l in levels:
            cll = CodeListLevel()
            cll.code_list = cl  # Point to the containing CodeList
            cll.code = l
            cll.description = None
            levels_dict[l] = cll
        # Codes
        codes_dict = create_dictionary()
        for ct in codes:
            c = Code()
            c.code = ct.code
            c.description = ct.description
            if ct.level in levels_dict:
                c.level = levels_dict[ct.level]  # Point to the containing CodeListLevel
            else:
                c.level = None
            codes_dict[ct.code] = c
            c.children = []
            c.parents = []
        # Set children & parents
        for ct in codes:
            for ch in ct.children:
                if ch in codes_dict:
                    c.children.append(codes_dict[ch])
                    codes_dict[ch].parents.append(c)

        return cl

    def clone(self) -> "CodeList":
        cl = CodeList()
        cl.code = self.code
        cl.description = self.description
        old_new_map = {}
        # Levels
        for cll1 in self.levels:
            cll = CodeListLevel()
            cll.code_list = cl
            cll.code = cll1.code
            # Codes in each level
            for ct1 in cll1.codes:
                c = Code()
                c.code = ct1.code
                c.description = ct1.description
                c.level = cll
                old_new_map[ct1] = c

        # Set children & parents
        for old, new in old_new_map.items():
            for ch in old.children:
                new.children.append(old_new_map[ch])
                old_new_map[ch].append(new)

        return cl


class CodeListLevel(ORMBase):
    __tablename__ = "dc_cl_levels"

    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(32))
    description = Column(String(1024))

    code_list_id = Column(Integer, ForeignKey(CodeList.id))
    code_list = relationship(CodeList, backref=backref("levels", cascade="all, delete-orphan"))


class Code(ORMBase): # A single value and its description for a Dimension
    __tablename__ = "dc_cl_codes"

    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(32))
    description = Column(String(1024))

    level_id = Column(Integer, ForeignKey(CodeListLevel.id))
    level = relationship(CodeListLevel, backref=backref("codes", cascade="all, delete-orphan"))

    geography = Column(Unicode)  # If it is a geographical code, what is the boundary
    geography_url = Column(String(1024))

    children = Column(Unicode)  # postgresql.JSONB)  # List of codes totalling this code
    parents = Column(Unicode)  # postgresql.JSONB)  # List of codes aggregating this code

    def __init__(self):
        self.children = []
        self.parents = []


@event.listens_for(Code, 'before_insert')
def code_before_insert(mapper, connection, target):
    target.children = json.dumps(target.children) if target.children else None
    target.parents = json.dumps(target.parents) if target.parents else None


@event.listens_for(Code, "load")
def code_after_load(target, context):
    target.parents = json.loads(target.parents) if target.parents else []
    target.children = json.loads(target.children) if target.children else []


class Concept(ORMBase):  # Concepts are independent of datasets
    __tablename__ = "dc_concepts"

    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(32))
    description = Column(String(1024))
    is_time = Column(Boolean, default=False)  # A concept having time nature

    attributes = Column(Unicode) # postgresql.JSONB)

    parent_id = Column(Integer, ForeignKey("dc_concepts.id"))
    parent = relationship("Concept", backref=backref("children", remote_side=[id]), cascade="all, delete-orphan")

    code_list_emb = Column(Unicode) # postgresql.JSONB)

    # Zero or one Code List
    code_list_id = Column(Integer, ForeignKey(CodeList.id))
    code_list = relationship(CodeList, backref=backref("concept", cascade="all, delete-orphan"))


@event.listens_for(Concept, 'before_insert')
def concept_before_insert(mapper, connection, target):
    target.code_list_emb = json.dumps(target.code_list_emb) if target.code_list_emb else None
    target.attributes = json.dumps(target.attributes) if target.attributes else None


@event.listens_for(Concept, "load")
def concept_after_load(target, context):
    target.attributes = json.loads(target.attributes) if target.attributes else {}
    target.code_list_emb = json.loads(target.code_list_emb) if target.code_list_emb else {}


# --------------------------------------------

class DataSource(ORMBase):  # Eurostat, OECD, FAO, ...
    __tablename__ = "dc_data_sources"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(32))
    description = Column(String(1024))
    data_dictionary = Column(Unicode) # postgresql.JSONB)  # "Metadata"


@event.listens_for(DataSource, 'before_insert')
def datasource_before_insert(mapper, connection, target):
    target.data_dictionary = json.dumps(target.data_dictionary) if target.data_dictionary else None


@event.listens_for(DataSource, "load")
def datasource_after_load(target, context):
    target.data_dictionary = json.loads(target.data_dictionary) if target.data_dictionary else {}


class Database(ORMBase): # A data source can have one or more databases
    __tablename__ = "dc_databases"

    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(32))
    description = Column(String(1024))
    data_dictionary = Column(Unicode) # postgresql.JSONB)  # "Metadata"

    data_source_id = Column(Integer, ForeignKey(DataSource.id))
    data_source = relationship(DataSource)


@event.listens_for(Database, 'before_insert')
def database_before_insert(mapper, connection, target):
    target.data_dictionary = json.dumps(target.data_dictionary) if target.data_dictionary else None


@event.listens_for(Database, "load")
def database_after_load(target, context):
    target.data_dictionary = json.loads(target.data_dictionary) if target.data_dictionary else {}


class Dataset(ORMBase):  # A database has many datasets
    __tablename__ = "dc_datasets"

    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(128))
    description = Column(String(1024))
    data_dictionary = Column(Unicode)  # postgresql.JSONB)  # "Metadata"

    attributes = Column(Unicode)  # postgresql.JSONB)

    database_id = Column(Integer, ForeignKey(Database.id))
    database = relationship(Database)

    # TODO Dimensions, Measures
    # TODO Stores

    def __init__(self):
        self.data = None  # type: pd.DataFrame

    def get_objects_list(self):
        """
        A list of objects related to the dataset is elaborated.
        This list is used for serialization/deserialization.
        :return:
        """
        lst = [self]
        if self.database:
            lst.append(self.database)
            if self.database.data_source:
                lst.append(self.database.data_source)
        # TODO Append "Concept" related to each dimension?
        for d in self.dimensions:
            lst.append(d)
            if d.code_list:
                lst.append(d.code_list)
                for l in d.code_list.levels:
                    lst.append(l)
                    for c in l.codes:
                        lst.append(c)
            # if d.get_hierarchy():
            #     # ADD: Hierarchy, HierarchyLevel(s), HierarchyNode(s), HierarchySource (if defined),
            #     #      HierarchyGroup (if defined). Referred HierarchyNode, Hierarchy
            #     h = d.get_hierarchy()
            #     lst.append(h)
            #     if h.hierarchy_group:
            #         lst.append(h.hierarchy_group)
            #         if h.hierarchy_group.hierarchy_source:
            #             lst.append(h.hierarchy_group.hierarchy_source)
            #     levels = set()
            #     other_hierarchies = set()
            #     for code in h.codes:
            #         lst.append(code)
            #         if code.referred_node:
            #             lst.append(code.referred_node)
            #             if code.referred_code.hierarchy not in other_hierarchies:
            #                 other_hierarchies.add(code.referred_code.hierarchy)
            #
            #         if code.level and code.level not in levels:
            #             levels.add(code.level)
            #
            #     for level in levels:
            #         lst.append(level)
            #     for hierarchy in other_hierarchies:
            #         lst.append(hierarchy)

        return lst

    @orm.reconstructor
    def init_on_load(self):
        self.data = None
        self.attributes = json.loads(self.attributes) if self.attributes else {}
        self.data_dictionary = json.loads(self.data_dictionary) if self.data_dictionary else {}


@event.listens_for(Dataset, 'before_insert')
def dataset_before_insert(mapper, connection, target):
    target.data_dictionary = json.dumps(target.data_dictionary) if target.data_dictionary else None
    target.attributes = json.dumps(target.attributes) if target.attributes else None


# Replaced by the "@orm.reconstructor" (equivalent)
# @event.listens_for(Dataset, "load")
# def dataset_after_load(target, context):
#     target.attributes = json.loads(target.attributes) if target.attributes else {}
#     target.data_dictionary = json.loads(target.data_dictionary) if target.data_dictionary else {}


class Dimension(ORMBase):  # A dimension is a concept linked to a dataset. It can be a Measure (a Measure is a kind of Dimension)
    __tablename__ = "dc_ds_dimensions"

    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(32))
    description = Column(String(1024))
    is_time = Column(Boolean, default=False)
    is_measure = Column(Boolean, default=False)
    attributes = Column(Unicode)  # postgresql.JSONB)

    dataset_id = Column(Integer, ForeignKey(Dataset.id))
    dataset = relationship(Dataset, lazy='subquery', backref=backref("dimensions", cascade="all, delete-orphan"))

    # Zero or one Concept
    # concept_id = Column(Integer, ForeignKey(Concept.id))  # The concept will contain the full code list
    # concept = relationship(Concept)

    code_list_emb = Column(Unicode)  # postgresql.JSONB)  # A reduced code list of the concept, if it exists
    # Zero or one Code List, this contains the reduced Code List
    code_list_id = Column(Integer, ForeignKey(CodeList.id))
    code_list = relationship(CodeList, backref=backref("dimension", cascade="all, delete-orphan"))

    def __init__(self):
        self.hierarchy = None  # type: Hierarchy

    @orm.reconstructor
    def init_on_load(self):
        self.attributes = json.loads(self.attributes) if self.attributes else {}
        self.code_list_emb = json.loads(self.code_list_emb) if self.code_list_emb else {}
        # Because "get_hierarchy()" method call the conversion in case of need there is no need of initialization here
        # It is an "on-demand" strategy
        self.hierarchy = None

    def get_hierarchy(self) -> Hierarchy:
        if not self.hierarchy and self.code_list:
            from nexinfosys.models.musiasem_concepts_helper import convert_code_list_to_hierarchy  # Local import to avoid circular imports (musiasem_concepts_helper imports statistical_datasets)
            self.hierarchy = convert_code_list_to_hierarchy(self.code_list)
        return self.hierarchy


@event.listens_for(Dimension, 'before_insert')
def dimension_before_insert(mapper, connection, target):
    target.code_list_emb = json.dumps(target.code_list_emb) if target.code_list_emb else None
    target.attributes = json.dumps(target.attributes) if target.attributes else None
    # TODO Convert Hierarchy to CodeList, if it makes sense (Hierarchy has more expresivity)


# Replaced by the "@orm.reconstructor" (equivalent)
# @event.listens_for(Dimension, "load")
# def dimension_after_load(target, context):
#     target.attributes = json.loads(target.attributes) if target.attributes else {}
#     target.code_list_emb = json.loads(target.code_list_emb) if target.code_list_emb else {}


class Store(ORMBase): # Location for one or more datasets. Datasets point to Stores, also a dataset can be in multiple locations
    __tablename__ = "dc_ds_stores"

    id = Column(Integer, primary_key=True, autoincrement=True)
    parameters = Column(Unicode) # postgresql.JSONB)

    dataset_id = Column(Integer, ForeignKey(Dataset.id))
    dataset = relationship(Dataset)


@event.listens_for(Store, 'before_insert')
def store_before_insert(mapper, connection, target):
    target.parameters = json.dumps(target.parameters) if target.parameters else None


@event.listens_for(Store, "load")
def store_after_load(target, context):
    target.parameters = json.loads(target.parameters) if target.parameters else {}


"""
Consider also caching and refresh policy
Consider dimensions defined in MuSIASEM, and mappings
"""



import logging

from sqlalchemy.orm import class_mapper
import pandas as pd
import numpy as np
import json
import sys
import blosc

# Some ideas from function "model_to_dict" (Google it, StackOverflow Q&A)
from nexinfosys.common.helper import PartialRetrievalDictionary, create_dictionary
from nexinfosys.models import MODEL_VERSION
from nexinfosys.models.musiasem_methodology_support import serialize_from_object, deserialize_to_object
from nexinfosys.model_services import State, get_case_study_registry_objects


def serialize(o_list):
    """
    Receives a list of SQLAlchemy objects to serialize
    The objects can be related between them by OneToMany and ManyToOne relations
    Returns a list of dictionaries with their properties and two special
    properties, "_nis_class_name" and "_nis_object_id" allowing the reconstruction

    Raise exception if some of the objects refers to an object OUT of the graph

    :param o_list:
    :return:
    """

    def fullname(o):
        return o.__module__ + "." + o.__class__.__name__

    # Dictionary to map obj to ID
    d_ref = {o: i for i, o in enumerate(o_list)}

    # Expand the list to referred objects
    cont = len(o_list)
    proc_lst = o_list
    while True:
        o_list2 = []
        for i, o in enumerate(proc_lst):
            # Relationships
            relationships = [(name, rel) for name, rel in class_mapper(o.__class__).relationships.items()]
            for name, relation in relationships:
                if str(relation.direction) != "symbol('ONETOMANY')":
                    ref_obj = o.__dict__.get(name)
                    if ref_obj:
                        if ref_obj not in d_ref:
                            d_ref[ref_obj] = cont
                            o_list2.append(ref_obj)
                            cont += 1

        o_list.extend(o_list2)
        proc_lst = o_list2
        if len(o_list2) == 0:
            break

    # Do the transformation to list of dictionaries
    d_list = []
    for i, o in enumerate(o_list):
        d = {c.key: getattr(o, c.key) for c in o.__table__.columns}
        d["_nis_class_name"] = fullname(o)
        d["_nis_object_id"] = i
        # Relationships
        relationships = [(name, rel) for name, rel in class_mapper(o.__class__).relationships.items()]
        for name, relation in relationships:
            if str(relation.direction) != "symbol('ONETOMANY')":
                ref_obj = o.__dict__.get(name)
                if ref_obj:
                    if ref_obj in d_ref:
                        d[name] = d_ref[ref_obj]
                else:
                    d[name] = -1  # None

        d_list.append(d)

    return d_list


def deserialize(d_list):
    """
    Receives a list of dictionaries representing SQLAlchemy object previously serialized

    :param d_list:
    :return:
    """
    def instantiate(full_class_name: str, c_dict: dict):
        import importlib
        if full_class_name in c_dict:
            class_ = c_dict[full_class_name]
        else:
            module_name, class_name = full_class_name.rsplit(".", 1)
            class_ = getattr(importlib.import_module(module_name), class_name)
            c_dict[full_class_name] = class_

        return class_()

    o_list = []
    c_list = {}  # Dictionary of classes (full_class_name, class)
    # Construct instances
    for d in d_list:
        o_list.append(instantiate(d["_nis_class_name"], c_list))
    # Now populate them
    ids = []
    for i, d in enumerate(d_list):
        o = o_list[i]
        o.__dict__.update({c.key: d[c.key] for c in o.__table__.columns})
        # Relationships
        relationships = [(name, rel) for name, rel in class_mapper(o.__class__).relationships.items()]
        for name, relation in relationships:
            if str(relation.direction) != "symbol('ONETOMANY')":
                ref_idx = d[name]
                if ref_idx < 0:
                    setattr(o, name, None)
                else:
                    setattr(o, name, o_list[ref_idx])
                    for t in relation.local_remote_pairs:
                        o_id_name = t[0].name  # Or t[0].key
                        r_id_name = t[1].name  # Or t[1].key
                        ids.append((o, o_id_name, o_list[ref_idx], r_id_name))

    for t in ids:
        k = getattr(t[2], t[3])
        if k:
            setattr(t[0], t[1], k)

    return o_list


def serialize_state(state: State):
    """
    Serialization prepared for a given organization of the state

    :return:
    """

    def serialize_dataframe(df):
        return df.to_json(orient="split", index=False), \
               json.dumps({i[0]: str(i[1]) for i in df.dtypes.to_dict().items()})
        # list(df.index.names), df.to_dict()

    logging.debug("  serialize_state IN")

    import copy
    # "_datasets"
    ns_ds = {}
    # Save and nullify "_datasets" before deep copy
    for ns in state.list_namespaces():
        _, _, _, datasets, _ = get_case_study_registry_objects(state, ns)
        ns_ds[ns] = datasets
        state.set("_datasets", create_dictionary(), ns)  # Nullify datasets

    # !!! WARNING: It destroys "state", so a DEEP COPY is performed !!!
    tmp = sys.getrecursionlimit()
    sys.setrecursionlimit(10000)
    state2 = copy.deepcopy(state)
    sys.setrecursionlimit(tmp)

    # Restore "_datasets"
    for ns in state.list_namespaces():
        state.set("_datasets", ns_ds[ns], ns)  # Nullify datasets

    # Iterate all namespaces
    for ns in state2.list_namespaces():
        glb_idx, p_sets, hh, _, mappings = get_case_study_registry_objects(state2, ns)
        if glb_idx:
            tmp = glb_idx.to_pickable()
            state2.set("_glb_idx", tmp, ns)
        lcia_methods = state2.get("_lcia_methods", ns)
        if lcia_methods:
            tmp = lcia_methods.to_pickable()
            state2.set("_lcia_methods", tmp, ns)
        _, _, _, datasets, _ = get_case_study_registry_objects(state, ns)
        datasets2 = create_dictionary()
        # TODO Serialize other DataFrames.
        # Process Datasets
        for ds_name in datasets:
            ds = datasets[ds_name]
            if isinstance(ds.data, pd.DataFrame):
                tmp = serialize_dataframe(ds.data)
            else:
                tmp = None
                # ds.data = None
            # DB serialize the datasets
            lst2 = serialize(ds.get_objects_list())
            lst2.append(tmp)  # Append the serialized DataFrame
            datasets2[ds_name] = lst2
        state2.set("_datasets", datasets2, ns)
    tmp = serialize_from_object(state2)  # <<<<<<<< SLOWEST !!!! (when debugging)
    logging.debug("  serialize_state length: "+str(len(tmp))+" OUT")
    tmp = blosc.compress(bytearray(tmp, "utf-8"), cname="zlib", typesize=8)
    logging.debug("  serialize_state compressed length: "+str(len(tmp))+" OUT")

    return tmp


def deserialize_state(st: str, state_version: int = MODEL_VERSION):
    """
    Deserializes an object previously serialized using "serialize_state"

    It can receive also a "State" modified for the serialization to restore it

    :param state_version: version number of the internal models
    :param st:
    :return:
    """
    def deserialize_dataframe(t):
        if t:
            dtypes = {i[0]: np.dtype(i[1]) for i in json.loads(t[1]).items()}
            df = pd.read_json(t[0], orient="split", dtype=dtypes)  # pd.DataFrame(t[1])
            # df.index.names = t[0]
            return df
        else:
            return pd.DataFrame()  # Return empty pd.Dataframe

    logging.debug("  deserialize_state")
    if isinstance(st, bytes):
        st = blosc.decompress(st).decode("utf-8")
    if isinstance(st, str):
        # TODO: use state_version to convert a previous version to the latest one
        #  This means transforming the old json to the latest json
        if state_version == MODEL_VERSION:
            state = deserialize_to_object(st)
        else:
            raise Exception(f"The model version {state_version} is not supported. Current version is {MODEL_VERSION}.")
    else:
        raise Exception(f"Serialized state must be a string: currently is of type {type(st)}")

    # Iterate all namespaces
    for ns in state.list_namespaces():
        lcia_methods = state.get("_lcia_methods", ns)
        if isinstance(lcia_methods, dict):
            lcia_methods = PartialRetrievalDictionary().from_pickable(lcia_methods)
            state.set("_lcia_methods", lcia_methods)
        glb_idx = state.get("_glb_idx", ns)
        if isinstance(glb_idx, dict):
            glb_idx = PartialRetrievalDictionary().from_pickable(glb_idx)
            state.set("_glb_idx", glb_idx)
        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state, ns)
        if isinstance(glb_idx, dict):
            logging.debug("glb_idx is DICT, after deserialization!!!")
        # TODO Deserialize DataFrames
        # In datasets
        for ds_name in datasets:
            lst = datasets[ds_name]
            ds = deserialize(lst[:-1])[0]
            ds.data = deserialize_dataframe(lst[-1])
            datasets[ds_name] = ds

    return state



from abc import abstractmethod

from typing import List

from nexinfosys.ie_imports.data_source_manager import IDataSourceManager
from nexinfosys.models.statistical_datasets import DataSource, Database, Dataset


class WIOD(IDataSourceManager):
    @abstractmethod
    def get_name(self) -> str:
        """ Source name """
        pass

    @abstractmethod
    def get_datasource(self) -> DataSource:
        """ Data source """
        pass

    @abstractmethod
    def get_databases(self) -> List[Database]:
        """ List of databases in the data source """
        pass

    @abstractmethod
    def get_datasets(self, database=None) -> list:
        """ List of datasets in a database, or in all the datasource (if database==None)
            Return a list of tuples (database, dataset)
        """
        pass

    @abstractmethod
    def get_dataset_structure(self, database, dataset) -> Dataset:
        """ Obtain the structure of a dataset: concepts, dimensions, attributes and measures """
        pass

    @abstractmethod
    def etl_full_database(self, database=None, update=False):
        """ If bulk download is supported, refresh full database """
        pass

    @abstractmethod
    def etl_dataset(self, dataset, update=False):
        """ If bulk download is supported, refresh full dataset """
        pass

    @abstractmethod
    def get_dataset_filtered(self, dataset, dataset_params: list) -> Dataset:
        """ Obtains the dataset with its structure plus the filtered values
            The values can be in a pd.DataFrame or in JSONStat compact format
            After this, new dimensions can be joined, aggregations need to be performed
        """
        pass

    @abstractmethod
    def get_refresh_policy(self):  # Refresh frequency for list of databases, list of datasets, and dataset
        pass



import collections

# SDMX Concept can be: dimension, attribute or measure. Stored in "metadatasets" associated to a dataset by its name
SDMXConcept = collections.namedtuple('Concept', 'type name istime description code_list')
